from textwrap import dedent

from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render
from django.urls import reverse

from core.decorators import web_require_scopes
from core.utils.settings_helper import send_email
from core.models import Notification, Organization, TeamInvitation, User
from core.types.htmx import HtmxHttpRequest


def delete_notification(user: User, code: TeamInvitation):
    notification = Notification.objects.filter(
        user=user,
        message="New Organization Invite",
        action="modal",
        action_value="accept_invite",
        extra_type="accept_invite_with_code",
        extra_value=code,
    ).first()

    if notification:
        notification.delete()


def check_team_invitation_is_valid(request, invitation: TeamInvitation, code=None):
    valid: bool = True

    if not invitation.is_active():
        valid = False
        messages.error(request, "Invitation has expired")

    if not valid:
        delete_notification(request.user, code)
        return False

    return True


@web_require_scopes("team:invite", True, True)
def send_user_team_invite(request: HtmxHttpRequest):
    user_email = request.POST.get("email")
    team_id = request.POST.get("team_id", "")
    team: Organization | None = Organization.objects.filter(leader=request.user, id=team_id).first()

    def return_error_notif(request: HtmxHttpRequest, message: str, autohide=None):
        messages.error(request, message)
        context = {"autohide": False} if autohide is False else {}
        resp = render(request, "partials/messages_list.html", context=context, status=200)
        resp["HX-Trigger-After-Swap"] = "invite_user_error"
        return resp

    if not user_email:
        return return_error_notif(request, "Please enter a valid user email")

    if not team:
        return return_error_notif(request, "You are not the leader of this team")

    user: User | None = User.objects.filter(email=user_email).first()

    if not user:
        return return_error_notif(request, 'User not found. Either ask them to create an account or press "Create User"')

    if user.teams_joined.filter(pk=team_id).exists():
        return return_error_notif(request, "User already is in this team")

    invitation = TeamInvitation.objects.create(team=team, user=user, invited_by=request.user)

    Notification.objects.create(
        user=user,
        message=f"New Organization Invite",
        action="modal",
        action_value="accept_invite",
        extra_type="accept_invite_with_code",
        extra_value=invitation.code,
    )

    send_email(
        destination=user.email,
        subject="New Organization Invite",
        content=dedent(
            f"""
                Hi {user.first_name or "User"},

                {request.user.first_name or f"User {request.user.email}"} has invited you to join the organization \"{team.name}\" (#{team.id})


                Click the url below to accept the invite!
                {request.build_absolute_uri(reverse("core:api:teams:join accept", kwargs={"code": invitation.code}))}

                Didn't give permission to be added to this organization? You can safely ignore the email, no actions can be done on
                behalf of you without your action.
            """
        ),
    )

    messages.success(request, "Invitation successfully sent")
    response = HttpResponse(status=200)
    response["HX-Refresh"] = "true"
    return response


def accept_team_invite(request: HtmxHttpRequest, code):
    invitation: TeamInvitation | None = TeamInvitation.objects.filter(code=code).prefetch_related("team").first()

    if not invitation:
        messages.error(request, "Invalid Invite Code")
        # Force break early to avoid "no invitation" on invitation.code
        delete_notification(request.user, code)
        return render(request, "partials/messages_list.html")

    if not check_team_invitation_is_valid(request, invitation, code):
        messages.error(request, "Invalid invite - Maybe it has expired?")
        return render(request, "partials/messages_list.html")

    if request.user.teams_joined.filter(pk=invitation.team_id).exists():
        messages.error(request, "You are already in this team")
        response = render(request, "partials/messages_list.html", status=200)
        response["HX-Trigger-After-Swap"] = "accept_invite_error"
        return response

    invitation.team.members.add(request.user)

    notification = Notification.objects.filter(
        user=request.user,
        action="modal",
        action_value="accept_invite",
        extra_type="accept_invite_with_code",
        extra_value=code,
    ).first()

    if notification:
        notification.delete()

    Notification.objects.create(
        user=request.user,
        message=f"You have now joined the team {invitation.team.name}",
        action="normal",
    )

    Notification.objects.create(
        user=invitation.invited_by,
        message=f"{request.user.username} has joined your team",
        action="normal",
    )

    invitation.delete()

    messages.success(request, f"You have successfully joined the team {invitation.team.name}")
    response = HttpResponse(status=200)
    response["HX-Refresh"] = "true"
    return response
    # return render(request, "partials/messages_list.html")


def decline_team_invite(request: HtmxHttpRequest, code):
    invitation: TeamInvitation | None = TeamInvitation.objects.filter(code=code).first()
    confirmation_text = request.POST.get("confirmation_text")

    if not invitation:
        messages.error(request, "Invalid Invite Code")
        # Force break early to avoid "no invitation" on invitation.code
        delete_notification(request.user, code)
        return render(request, "partials/messages_list.html")

    if not check_team_invitation_is_valid(request, invitation, code):
        return render(request, "partials/messages_list.html")

    # if confirmation_text != "i confirm i want to decline " + invitation.team.name:
    #     messages.error(request, "Invalid confirmation text")
    # return redirect("teams:dashboard join", code=code)  # kwargs={"code": code})

    invitation.team.members.remove(request.user)

    Notification.objects.create(
        user=request.user,
        message=f"You have declined the team invitation",
        action="normal",
    )

    Notification.objects.create(
        user=invitation.invited_by,
        message=f"{request.user.username} has declined the team invitation",
        action="normal",
    )

    delete_notification(request.user, code)

    invitation.delete()
    messages.success(request, "You have successfully declined the team invitation")

    return render(request, "partials/messages_list.html")



from typing import Any

from django import template

register = template.Library()


@register.simple_tag
def dict_get(dictionary: dict, key: Any):
    return dictionary.get(key)


register.filter("dict_get", dict_get)



from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, render
from django.urls import reverse

from billing.service.entitlements import has_entitlement, get_entitlements


def has_entitlements_called_from_backend_handler(entitlements: list[str] | str, htmx_api: bool = False):
    def decorator(view_func):
        def wrapper_func(request, *args, **kwargs):
            user_does_have_entitlements: bool
            if isinstance(entitlements, (list, set)):
                users_entitlements = get_entitlements(request.actor)
                user_does_have_entitlements = all(entitlement in users_entitlements for entitlement in entitlements)
            else:
                user_does_have_entitlements = has_entitlement(request.actor, entitlements)

            if user_does_have_entitlements:
                return view_func(request, *args, **kwargs)
            else:
                messages.warning(request, f"Your plan unfortunately doesn't include this feature.")

                if htmx_api:
                    return render(request, "base/toast.html", {"autohide": False})
                elif request.htmx:
                    return HttpResponseRedirect(reverse("billing:dashboard"))
                return redirect("billing:dashboard")

        return wrapper_func

    return decorator



from core.models import UserSettings
from PIL import Image

from core.utils.dataclasses import BaseServiceResponse


class UpdateProfilePictureServiceResponse(BaseServiceResponse[str]): ...


def update_profile_picture(profile_picture, user_profile: UserSettings) -> UpdateProfilePictureServiceResponse:
    if not profile_picture:
        return UpdateProfilePictureServiceResponse(error_message="Invalid or unsupported image file")

    try:
        # Max file size is 10MB (Change the first number to determine the size in MB)
        max_file_size = 10 * 1024 * 1024

        if profile_picture.size is None:
            return UpdateProfilePictureServiceResponse(error_message="File size not found")

        if profile_picture.size > max_file_size:
            return UpdateProfilePictureServiceResponse(error_message="File size should be up to 10MB")

        img = Image.open(profile_picture)
        img.verify()

        if img.format is None or img.format.lower() not in ["jpeg", "png", "jpg"]:
            return UpdateProfilePictureServiceResponse(
                error_message="Unsupported image format. We support only JPEG, JPG, PNG, if you have a good extension, your file just got renamed."
            )

        user_profile.profile_picture = profile_picture
        user_profile.save()
        return UpdateProfilePictureServiceResponse(True, "Successfully updated profile picture")
    except (FileNotFoundError, Image.UnidentifiedImageError):
        return UpdateProfilePictureServiceResponse(error_message="Invalid or unsupported image file")



from django.contrib import messages
from django.contrib.auth.hashers import check_password
from django.http import (
    HttpRequest,
)
from django.shortcuts import redirect
from django.utils import timezone
from django.views.decorators.http import require_POST

from core.decorators import not_authenticated
from core.models import PasswordSecret


@not_authenticated
@require_POST
def set_password_set(request: HttpRequest, secret):
    password = request.POST.get("password", "")
    if len(password) > 7:
        SECRET_RETURNED = PasswordSecret.objects.all()

        for SECRET in SECRET_RETURNED:
            if SECRET.expires is not None and SECRET.expires < timezone.now():
                SECRET.delete()
                continue
            elif check_password(secret, SECRET.secret):
                USER = SECRET.user
                USER.set_password(password)
                USER.save()
                SECRET.delete()
                messages.success(request, "Successfully changed your password.")
                return redirect("core:auth:login")

        messages.error(
            request,
            "Invalid password code. The code has either expired or was not entered correctly."
            "Please contact an administrator for support.",
        )
        return redirect("core:auth:login")

    else:
        messages.error(request, "No code provided. Please contact an administrator for support.")
        return redirect("core:auth:login")

    messages.error(request, "Sorry, somethging went wrong!")
    return redirect("core:auth:login")



import time

from django import template

register = template.Library()


def split(string, char=" "):
    return string.split(char)


def dashify(string, recurrence=2):
    num_str = str(string)

    return "-".join(num_str[i : i + recurrence] for i in range(0, len(num_str), recurrence))


def to_list(string, separator=",") -> list[str]:
    return string.split(separator)


def contains(value, arg):
    return arg in str(value)


def day_to_number_sunday(day: str) -> int:
    """
    Converts a day of the week to a number with Sunday as the first day.

    Args:
        day (str): The day of the week (e.g., "Sunday", "Monday").

    Returns:
        int: The corresponding number with Sunday as 0 and Saturday as 6.
    """
    # Get the day number with Monday as 0
    day_number = time.strptime(day, "%A").tm_wday

    # Adjust the day number to make Sunday the first day
    sunday_first_day_number = (day_number + 1) % 7

    return sunday_first_day_number


def day_to_number_monday(day: str) -> int:
    return time.strptime(day, "%A").tm_wday + 1


def month_to_number(month: str) -> int:
    return time.strptime(month, "%B").tm_mon


register.filter("split", split)
register.filter("dashify", dashify)
register.filter("contains", contains)
register.filter("to_list", to_list)
register.filter("day_to_number_monday", day_to_number_monday)
register.filter("day_to_number_sunday", day_to_number_sunday)
register.filter("month_to_number", month_to_number)



from core.models import User, Organization
from core.service.api_keys.get import get_api_key_by_name
from core.api.public import APIAuthToken


def delete_api_key(request, owner: User | Organization, key: str | None | APIAuthToken) -> bool | str:
    if isinstance(owner, Organization) and "api_keys:write" not in owner.permissions.get(user=request.user).scopes:
        return "No permission to delete key"

    if not isinstance(key, APIAuthToken):
        key: APIAuthToken | None = get_api_key_by_name(owner, key)  # type: ignore[no-redef, arg-type]

    if not key:
        return "Key not found"

    key.deactivate()  # type: ignore[union-attr]

    return True



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin

from core.api.public import APIAuthToken

from core.models import (
    PasswordSecret,
    AuditLog,
    LoginLog,
    Error,
    TracebackError,
    UserSettings,
    Notification,
    Organization,
    TeamInvitation,
    TeamMemberPermission,
    User,
    FeatureFlags,
    VerificationCodes,
    EmailSendStatus,
)

from django.conf import settings

# from django.contrib.auth.models imp/ort User
# admin.register(Invoice)
admin.site.register(
    [
        UserSettings,
        PasswordSecret,
        AuditLog,
        LoginLog,
        Error,
        TracebackError,
        Notification,
        Organization,
        TeamInvitation,
        TeamMemberPermission,
        FeatureFlags,
        VerificationCodes,
        APIAuthToken,
    ]
)

if getattr(settings, "BILLING_ENABLED", False):
    from billing.models import PlanFeature, PlanFeatureGroup, SubscriptionPlan, UserSubscription

    admin.site.register([PlanFeature, PlanFeatureGroup, SubscriptionPlan, UserSubscription])


class EmailSendStatusAdmin(admin.ModelAdmin):
    readonly_fields = ["aws_message_id"]


admin.site.register(EmailSendStatus, EmailSendStatusAdmin)

# admin.site.unregister(User)
fields = list(UserAdmin.fieldsets)  # type: ignore[arg-type]
fields[0] = (
    None,
    {
        "fields": (
            "username",
            "password",
            "logged_in_as_team",
            "awaiting_email_verification",
            "stripe_customer_id",
            "entitlements",
            "require_change_password",
        )
    },
)
UserAdmin.fieldsets = tuple(fields)
admin.site.register(User, UserAdmin)

admin.site.site_header = "Strelix Core Admin"
admin.site.index_title = "Strelix"
admin.site.site_title = "Strelix | Administration"



from uuid import uuid4

from django.db import models

from core.models import OwnerBase

from django.utils import timezone

from django.utils.timezone import now as timezone_now


class SubscriptionPlan(models.Model):
    """
    Subscription plans available for users.
    """

    name = models.CharField(max_length=50, unique=True)
    price_per_month = models.DecimalField(max_digits=10, decimal_places=2, null=True, blank=True)
    description = models.TextField(max_length=500, null=True, blank=True)
    stripe_product_id = models.CharField(max_length=100, null=True, blank=True)
    stripe_price_id = models.CharField(max_length=100, null=True, blank=True)


def __str__(self):
    return f"{self.name} - {self.price_per_month or 'free' if self.price_per_month != -1 else 'custom'}"


class UserSubscription(OwnerBase):
    """
    Track which subscription plan a user is currently subscribed to.
    """

    uuid = models.UUIDField(unique=True, default=uuid4)
    subscription_plan = models.ForeignKey(SubscriptionPlan, on_delete=models.SET_NULL, null=True)
    stripe_subscription_id = models.CharField(max_length=100, null=True, blank=True)
    # Custom price only used for enterprise or negotiated plans
    custom_subscription_price_per_month = models.DecimalField(max_digits=10, decimal_places=2, null=True, blank=True)

    start_date = models.DateTimeField(auto_now_add=True)
    end_date = models.DateTimeField(null=True, blank=True)

    def __str__(self):
        return f"{self.owner} - {self.subscription_plan.name} ({self.start_date} to {self.end_date or 'ongoing'})"

    @property
    def has_ended(self):
        return bool(self.end_date)

    @property
    def get_price(self):
        return self.custom_subscription_price_per_month or self.subscription_plan.price_per_month or "0.00"

    def end_now(self):
        self.end_date = timezone.now()
        self.save()
        return self


class PlanFeatureGroup(models.Model):
    name = models.CharField(max_length=50)  # E.g. 'invoices'


class PlanFeature(models.Model):
    """
    Details related to certain features. E.g. "emails sent", we can allow site admins to change prices, units, and customise their
    billing
    """

    slug = models.CharField(max_length=100)
    stripe_price_id = models.CharField(max_length=100, null=True, blank=True)
    description = models.TextField(max_length=500, null=True, blank=True)

    max_limit_per_month = models.IntegerField(null=True, blank=True)

    subscription_plan = models.ForeignKey(SubscriptionPlan, on_delete=models.CASCADE, related_name="features")
    group = models.ForeignKey(PlanFeatureGroup, on_delete=models.CASCADE, related_name="features")

    def __str__(self):
        return f"{self.slug} - subscription id: {self.subscription_plan_id}"


class StripeWebhookEvent(models.Model):
    event_id = models.CharField(max_length=100, unique=True)
    event_type = models.CharField(max_length=100)  # e.g. 'customer.subscription.created'
    data = models.JSONField()
    raw_event = models.JSONField()


class StripeCheckoutSession(OwnerBase):
    uuid = models.UUIDField(unique=True, default=uuid4)

    stripe_session_id = models.CharField(max_length=100, unique=True, blank=True, null=True)
    plan = models.ForeignKey(SubscriptionPlan, on_delete=models.CASCADE, related_name="checkout_sessions")
    features = models.ManyToManyField(PlanFeature, related_name="checkout_sessions")


class BillingUsage(OwnerBase):
    EVENT_TYPES = (
        ("usage", "Metered Usage"),
        # ("storage", "Storage"),
    )

    event_name = models.CharField(max_length=100)  # e.g. 'invoices-created'
    event_type = models.CharField(max_length=20, choices=EVENT_TYPES, default="usage")
    quantity = models.PositiveSmallIntegerField(default=1)  # e.g. 1

    created_at = models.DateTimeField(auto_now_add=True)

    processed_at = models.DateTimeField(null=True, blank=True)
    processed = models.BooleanField(default=False)
    stripe_unique_usage_identifier = models.CharField(max_length=100, null=True, blank=True)

    def set_processed(self, processed_time):
        self.processed = True
        self.processed_at = processed_time or timezone_now()
        self.save()
        return self



from django.shortcuts import render

from core.decorators import web_require_scopes
from billing.models import UserSubscription, SubscriptionPlan
from core.types.requests import WebRequest


@web_require_scopes("billing:manage", api=True, htmx=True)
def billing_dashboard_endpoint(request: WebRequest):
    context: dict = {}

    subscriptions = UserSubscription.filter_by_owner(request.actor).select_related("subscription_plan").all()
    all_subscription_plans = SubscriptionPlan.objects.all()

    if subscriptions.exists():
        context["free_plan_available"] = True

    context.update(
        {
            "active_subscription": subscriptions.filter(end_date__isnull=True).first(),
            "all_user_subscriptions": subscriptions,
            "all_subscription_plans": all_subscription_plans,
        }
    )

    return render(
        request,
        "pages/billing/dashboard/dashboard.html",
        context,
    )


@web_require_scopes("billing:manage", api=True, htmx=True)
def all_subscriptions_htmx_endpoint(request: WebRequest):
    context: dict = {}

    subscriptions = UserSubscription.filter_by_owner(request.actor).select_related("subscription_plan").all()

    context.update(
        {
            "active_subscription": subscriptions.filter(end_date__isnull=True).first(),
            "all_user_subscriptions": subscriptions,
        }
    )

    return render(
        request,
        "pages/billing/dashboard/all_subscriptions.html",
        context,
    )



from django.core.exceptions import ValidationError

from core.api.public import APIAuthToken
from core.models import User, Organization
from core.service.permissions.scopes import validate_scopes


def generate_public_api_key(
    request,
    owner: User | Organization,
    api_key_name: str | None,
    permissions: list,
    *,
    expires=None,
    description=None,
    administrator_toggle: bool = False,
    administrator_type: str | None = None,
) -> tuple[APIAuthToken | None, str]:
    if not validate_name(api_key_name):
        return None, "Invalid key name"

    if not validate_description(description):
        return None, "Invalid description"

    if api_key_exists_under_name(owner, api_key_name):
        return None, "A key with this name already exists in your account"

    if err_validate_scopes := validate_scopes(permissions):  # or not has_permission_to_create(request, owner):
        if err_validate_scopes.failed:
            return None, err_validate_scopes.error

    administrator_service_type = None

    if request.user.is_superuser:
        if administrator_toggle:
            if administrator_type not in [option[0] for option in APIAuthToken.AdministratorServiceTypes.choices]:
                return None, "Invalid administration type"
            administrator_service_type = administrator_type

    token = APIAuthToken(
        name=api_key_name,
        description=description,
        expires=expires,
        scopes=permissions,
        administrator_service_type=administrator_service_type,
    )  # type: ignore[arg-type, misc]

    raw_key: str = token.generate_key()

    if isinstance(owner, Organization):
        token.organization = owner
    else:
        token.user = owner

    try:
        token.full_clean()
    except ValidationError as validation_errors:
        field, error_list = next(iter(validation_errors.error_dict.items()))

        field = "Permissions" if field == "scopes" else field.title()

        if isinstance(error_list[0], ValidationError):
            error_message = error_list[0].messages[0]
        else:
            error_message = error_list[0]

        return None, f"{field}: {error_message}"

    token.save()

    return token, raw_key


def validate_name(name: str | None) -> bool:
    """
    Require name not already exist under account
    """
    if not name:
        return False
    return len(name) <= 64


def validate_description(description: str | None) -> bool:
    """
    Accept any description
    Reject description longer than 255 characters
    """
    return not description or len(description) <= 255


def api_key_exists_under_name(owner: User | Organization, name: str | None) -> bool:
    """
    Check if API key exists under a given name
    """
    return APIAuthToken.filter_by_owner(owner).filter(name=name, active=True).exists()


def has_permission_to_create(request, owner: User | Organization) -> bool:
    if isinstance(owner, User):
        return True

    if owner.permissions.filter(user=request.user).exists() and "api_keys:write" in owner.permissions.get(user=request.user).scopes:
        return True
    return False



from __future__ import annotations

import logging
from functools import wraps
from typing import TypedDict

from django.contrib import messages
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.shortcuts import render
from django.urls import reverse

from core.models import TeamMemberPermission
from core.types.requests import WebRequest
from core.utils.feature_flags import get_feature_status

logger = logging.getLogger(__name__)


def not_authenticated(view_func):
    def wrapper_func(request, *args, **kwargs):
        if request.user.is_authenticated:
            return redirect("dashboard")
        else:
            return view_func(request, *args, **kwargs)

    return wrapper_func


def staff_only(view_func):
    def wrapper_func(request, *args, **kwargs):
        if request.user.is_staff and request.user.is_authenticated:
            return view_func(request, *args, **kwargs)
        else:
            messages.error(request, "You don't have permission to view this page.")
            return redirect("dashboard")

    return wrapper_func


def superuser_only(view_func):
    def wrapper_func(request, *args, **kwargs):
        if request.user.is_authenticated and request.user.is_superuser:
            return view_func(request, *args, **kwargs)
        else:
            messages.error(request, "You don't have permission to view this page.")
            return redirect("dashboard")

    return wrapper_func


def htmx_only(viewname: str = "dashboard"):
    def decorator(view_func):
        def wrapper_func(request, *args, **kwargs):
            if request.htmx:
                return view_func(request, *args, **kwargs)
            else:
                return redirect(viewname)

        return wrapper_func

    return decorator


def hx_boost(view):
    """
    Decorator for HTMX requests.

    used by wrapping FBV in @hx_boost and adding **kwargs to param
    then you can use context = kwargs.get("context", {}) to continue and then it will handle HTMX boosts
    """

    @wraps(view)
    def wrapper(request, *args, **kwargs):
        if request.htmx.boosted:
            kwargs["context"] = kwargs.get("context", {}) | {"base": "core/base/htmx.html"}
        return view(request, *args, **kwargs)

    return wrapper


def feature_flag_check(flag, status=True, api=False, htmx=False):
    def decorator(view_func):
        @wraps(view_func)
        def wrapper(request, *args, **kwargs):
            feat_status = get_feature_status(flag)

            if feat_status == status:
                return view_func(request, *args, **kwargs)

            if api and htmx:
                messages.error(request, "This feature is currently disabled.")
                return render(request, "core/base/toasts.html")
            elif api:
                return HttpResponse(status=403, content="This feature is currently disabled.")
            messages.error(request, "This feature is currently disabled.")
            try:
                last_visited_url = request.session["last_visited"]
                current_url = request.build_absolute_uri()
                if last_visited_url != current_url:
                    return HttpResponseRedirect(last_visited_url)
            except KeyError:
                pass
            return HttpResponseRedirect(reverse("dashboard"))

        return wrapper

    return decorator


class FlagItem(TypedDict):
    name: str
    desired: bool


def feature_flag_check_multi(flag_list: list[FlagItem], api=False, htmx=False):
    """
    Checks if at least one of the flags in the list is the desired status
    """

    def decorator(view_func):
        @wraps(view_func)
        def wrapper(request, *args, **kwargs):
            if not any(get_feature_status(flag["name"]) == flag["desired"] for flag in flag_list):
                if api and htmx:
                    messages.error(request, "This feature is currently disabled.")
                    return render(request, "core/base/toasts.html")
                elif api:
                    return HttpResponse(status=403, content="This feature is currently disabled.")
                messages.error(request, "This feature is currently disabled.")
                return HttpResponseRedirect(request.META.get("HTTP_REFERER"))

            return view_func(request, *args, **kwargs)

        return wrapper

    return decorator


# def quota_usage_check(limit: str | QuotaLimit, extra_data: str | int | None = None, api=False, htmx=False):
#     def decorator(view_func):
#         @wraps(view_func)
#         def wrapper(request, *args, **kwargs):
#             try:
#                 quota_limit = QuotaLimit.objects.get(slug=limit) if isinstance(limit, str) else limit
#             except QuotaLimit.DoesNotExist:
#                 return view_func(request, *args, **kwargs)
#
#             if not quota_limit.strict_goes_above_limit(request.user, extra=extra_data):
#                 return view_func(request, *args, **kwargs)
#
#             if api and htmx:
#                 messages.error(request, f"You have reached the quota limit for this service '{quota_limit.slug}'")
#                 return render(request, "partials/messages_list.html", {"autohide": False})
#             elif api:
#                 return HttpResponse(status=403, content=f"You have reached the quota limit for this service '{quota_limit.slug}'")
#             messages.error(request, f"You have reached the quota limit for this service '{quota_limit.slug}'")
#             try:
#                 last_visited_url = request.session["last_visited"]
#                 current_url = request.build_absolute_uri()
#                 if last_visited_url != current_url:
#                     return HttpResponseRedirect(last_visited_url)
#             except KeyError:
#                 pass
#             return HttpResponseRedirect(reverse("dashboard"))
#
#         return wrapper
#
#     return decorator


not_logged_in = not_authenticated
logged_out = not_authenticated


def web_require_scopes(scopes: str | list[str], htmx=False, api=False, redirect_url=None):
    """
    Only to be used by WebRequests (htmx or html) NOT PUBLIC API
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request: WebRequest, *args, **kwargs):
            if request.team_id and not request.team:
                return return_error(request, "Team not found")

            if request.team:
                # Check for team permissions based on team_id and scopes
                if not request.team.is_owner(request.user):
                    team_permissions = TeamMemberPermission.objects.filter(team=request.team, user=request.user).first()

                    if not team_permissions:
                        return return_error(request, "You do not have permission to perform this action (no permissions for team)")

                    # single scope
                    if isinstance(scopes, str) and scopes not in team_permissions.scopes:
                        return return_error(request, f"You do not have permission to perform this action ({scopes})")

                    # scope list
                    if isinstance(scopes, list):
                        for scope in scopes:
                            if scope not in team_permissions.scopes:
                                return return_error(request, f"You do not have permission to perform this action ({scope})")
            return view_func(request, *args, **kwargs)

        _wrapped_view.required_scopes = scopes
        return _wrapped_view

    def return_error(request: WebRequest, msg: str):
        logging.info(f"User does not have permission to perform this action (User ID: {request.user.id}, Scopes: {scopes})")
        if api and htmx:
            messages.error(request, msg)
            return render(request, "core/base/toast.html", {"autohide": False})
        elif api:
            return HttpResponse(status=403, content=msg)
        elif request.htmx:
            messages.error(request, msg)
            resp = HttpResponse(status=200)

            try:
                last_visited_url = request.session["last_visited"]
                current_url = request.build_absolute_uri()
                if last_visited_url != current_url:
                    resp["HX-Replace-Url"] = last_visited_url
            except KeyError:
                ...
            resp["HX-Refresh"] = "true"
            return resp

        messages.error(request, msg)

        try:
            last_visited_url = request.session["last_visited"]
            current_url = request.build_absolute_uri()
            if last_visited_url != current_url:
                return HttpResponseRedirect(last_visited_url)
        except KeyError:
            pass

        if not redirect_url:
            return HttpResponseRedirect(reverse("dashboard"))

        try:
            return HttpResponseRedirect(reverse(redirect_url))
        except KeyError:
            return HttpResponseRedirect(reverse("dashboard"))

    return decorator


# wrapper around billing has_entitlements only load

from django.conf import settings


def has_entitlements(entitlements: list[str] | str, htmx_api: bool = False):
    def decorator(view_func):
        @wraps(view_func)
        def wrapper(request, *args, **kwargs):
            if settings.BILLING_ENABLED:
                from billing.decorators import has_entitlements_called_from_backend_handler

                wrapped_view_func = has_entitlements_called_from_backend_handler(
                    entitlements if isinstance(entitlements, list) else [entitlements], htmx_api
                )(view_func)
                return wrapped_view_func(request, *args, **kwargs)
            return view_func(request, *args, **kwargs)

        return wrapper

    return decorator



from typing import Optional, Any

from django.http import HttpRequest
from django.urls import resolve, reverse
from django.urls.exceptions import NoReverseMatch

ALL_ITEMS: dict[str, tuple[str, Optional[str], Optional[str]]] = {
    "dashboard": ("Dashboard", "dashboard", "house"),
    "finance:invoices:dashboard": ("Invoices", "finance:invoices:single:dashboard", "file-invoice"),
    "finance:invoices:single:dashboard": ("Single", "finance:invoices:single:dashboard", "file-invoice"),
    "finance:invoices:single:create": ("Create (single)", "finance:invoices:single:create", None),
    "finance:invoices:recurring:dashboard": ("Recurring", "finance:invoices:recurring:dashboard", "refresh"),
    "finance:invoices:recurring:create": ("Create (recurring)", "finance:invoices:recurring:create", None),
    "finance:invoices:single:edit": ("Edit", None, "pencil"),
    "finance:invoices:single:overview": ("Invoice", None, None),
    "receipts dashboard": ("Receipts", "receipts dashboard", "file-invoice"),
    "core:teams:dashboard": ("Teams", "core:teams:dashboard", "users"),
    "settings:dashboard": ("Settings", "settings:dashboard", "gear"),
    "clients:dashboard": ("Clients", "clients:dashboard", "users"),
    "clients:create": ("Create", "clients:create", None),
    "reports:dashboard": ("Monthly Reports", "reports:dashboard", "chart-line"),
}

ALL_BREADCRUMBS: dict[str, str | tuple] = {
    "dashboard": "dashboard",
    "core:teams:dashboard": ("dashboard", "core:teams:dashboard"),
    "receipts dashboard": ("dashboard", "receipts dashboard"),
    "finance:invoices:single:dashboard": ("dashboard", "finance:invoices:dashboard", "finance:invoices:single:dashboard"),
    "finance:invoices:single:create": ("dashboard", "finance:invoices:dashboard", "finance:invoices:single:create"),
    "finance:invoices:recurring:dashboard": ("dashboard", "finance:invoices:dashboard", "finance:invoices:recurring:dashboard"),
    "finance:invoices:recurring:create": ("dashboard", "finance:invoices:dashboard", "finance:invoices:recurring:create"),
    "finance:invoices:single:edit": ("dashboard", "finance:invoices:dashboard", "finance:invoices:single:edit"),
    "finance:invoices:single:overview": ("dashboard", "finance:invoices:dashboard", "finance:invoices:single:overview"),
    "clients:dashboard": ("dashboard", "clients:dashboard"),
    "clients:create": ("dashboard", "clients:dashboard", "clients:create"),
    "settings:dashboard": ("dashboard", "settings:dashboard"),
    "reports:dashboard": ("dashboard", "reports:dashboard"),
}


def get_item(name: str, url_name: Optional[str] = None, icon: Optional[str] = None, kwargs: dict = {}, *, request=None) -> dict:
    """
    Create a breadcrumb item dictionary.
    Parameters:
    - name (str): The name of the breadcrumb item.
    - url_name (str): The URL name used for generating the URL using Django's reverse function.
    - icon (Optional[str]): The icon associated with the breadcrumb item (default is None).
    Returns:
    Dict[str, Any]: A dictionary representing the breadcrumb item.
    """

    if request:
        rev_kwargs = {kwarg: request.resolver_match.kwargs.get(kwarg) for url, kwarg in kwargs.items() if url == url_name if kwargs}
    else:
        rev_kwargs = {}
    return {
        "name": name,
        "url": reverse(url_name, kwargs=rev_kwargs if rev_kwargs else {}) if url_name else "",
        "icon": icon,
    }


def generate_breadcrumbs(*breadcrumb_list: str, request=None) -> list[dict[Any, Any] | None]:
    """
    Generate a list of breadcrumb items based on the provided list of breadcrumb names.
    Parameters:
    - breadcrumb_list (str): Variable number of strings representing the names of the breadcrumbs.
    Returns:
    List[Dict[str, Any]]: A list of dictionaries representing the breadcrumb items.
    """
    return [
        get_item(*ALL_ITEMS.get(breadcrumb, (None, None, None)), request=request)
        for breadcrumb in breadcrumb_list
        if breadcrumb in ALL_ITEMS
    ]


def get_breadcrumbs(*, request: HttpRequest | None = None, url: str | None = None):
    current_url_name: str | Any = request.resolver_match.view_name if request and request.resolver_match else None  # type: ignore[ union-attr]
    if url:
        try:
            current_url_name = resolve(url).view_name
        except NoReverseMatch:
            return {"breadcrumb": []}
    return {"breadcrumb": generate_breadcrumbs(*ALL_BREADCRUMBS.get(current_url_name, []), request=request)}



from django.db.models import Q
from django.shortcuts import render, redirect

from core.models import QuotaLimit
from core.types.htmx import HtmxHttpRequest


def fetch_all_quotas(request: HtmxHttpRequest, group: str):
    context = {}
    if not request.htmx:
        return redirect("quotas")

    search_text = request.GET.get("search")

    results = QuotaLimit.objects.filter(slug__startswith=group).prefetch_related("quota_overrides", "quota_usage").order_by("-slug")

    if search_text:
        results = results.filter(Q(name__icontains=search_text))

    quotas = [
        {
            "quota_limit": ql.get_quota_limit(request.user),
            "period_usage": ql.get_period_usage(request.user),
            "quota_object": ql,
        }
        for ql in results
    ]

    context.update({"quotas": quotas})
    return render(request, "pages/quotas/_fetch_body.html", context)



from core.api.public import APIAuthToken
from core.types.requests import WebRequest
from core.utils.dataclasses import BaseServiceResponse


class APIAuthenticationServiceResponse(BaseServiceResponse[None]):
    response: None = None
    _status_code: int


def authenticate_api_key(request: WebRequest) -> APIAuthenticationServiceResponse:
    auth_header = request.headers.get("Authorization")

    if not (auth_header and auth_header.startswith("Bearer ")):
        return APIAuthenticationServiceResponse(error_message="Unauthorized", status_code=401)

    token_key = auth_header.split(" ")[1]

    try:
        token = APIAuthToken.objects.get(
            hashed_key=APIAuthToken.hash_raw_key(token_key),
            active=True,
            administrator_service_type=APIAuthToken.AdministratorServiceTypes.AWS_WEBHOOK_CALLBACK,
        )

        if token.has_expired:
            return APIAuthenticationServiceResponse(error_message="Token expired", status_code=400)
    except APIAuthToken.DoesNotExist:
        return APIAuthenticationServiceResponse(error_message="Token not found", status_code=400)

    token.update_last_used()

    return APIAuthenticationServiceResponse(True, None, status_code=200)



import os
import json
import uuid
import threading
import boto3
import inspect

from django.urls import reverse


class Task:
    def __init__(self, queue_url=None):
        self.queue_url: str | None = queue_url or os.environ.get("AWS_SQS_QUEUE_URL")
        self.region_name = os.environ.get("AWS_REGION_NAME")
        self.aws_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID")
        self.aws_secret_access_key = os.environ.get("AWS_ACCESS_KEY")
        self.WEBHOOK_URL = os.environ.get("SITE_URL", default="http://127.0.0.1:8000") + reverse("api:public:webhooks:receive_global")

        if self.queue_url:
            self.sqs_client = boto3.client(
                "sqs",
                # aws_access_key_id=self.aws_access_key_id,
                # aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name,
            )
        else:
            self.sqs_client = None  # Only set up SQS client if queue_url is provided

    def queue_task(self, func, *args, **kwargs):
        # Determine if func is a string or callable
        if isinstance(func, str):
            func_name = func
        elif callable(func):
            func_name = self._get_function_path(func)
        else:
            raise ValueError("func must be a callable or a string representing the function name.")

        # If SQS is not configured, execute the function directly
        if not self.queue_url or not self.region_name:
            return self.execute_now(func_name, *args, **kwargs)

        print("TASK 3 - call _send_message to SQS")
        # Use threading to send the message to SQS
        thread = threading.Thread(target=self._send_message, args=(func_name, args, kwargs))
        thread.start()

        return "Task submitted to SQS"

    def _send_message(self, func_name, args, kwargs):
        message_body = {"func_name": func_name, "args": args, "kwargs": kwargs, "webhook_url": self.WEBHOOK_URL}
        print(message_body)

        try:
            print("TASK 4 - Send SQS message")
            res = self.sqs_client.send_message(
                QueueUrl=self.queue_url,
                MessageBody=json.dumps(message_body),
                MessageGroupId="default",
                MessageDeduplicationId=str(uuid.uuid4()),
            )
            print(f"SQS Response: {res}")
        except Exception as e:
            print(f"Error sending message to SQS: {str(e)}")
            return self.execute_now(func_name, *args, **kwargs)

    def execute_now(self, func_name, *args, **kwargs):
        print("TASK 6 - Execute function directly")
        if "." not in func_name:
            raise ValueError(f"Invalid function name: {func_name}. Must include module path.")

        func = globals().get(func_name)
        if func is None:
            try:
                module_name, func_name = func_name.rsplit(".", 1)
                module = __import__(module_name, fromlist=[func_name])
                func = getattr(module, func_name)
            except (ImportError, AttributeError) as e:
                raise ValueError(f"Function {func_name} not found: {str(e)}")

        if callable(func):
            return func(*args, **kwargs)
        else:
            raise ValueError(f"Function {func_name} is not callable")

    def _get_function_path(self, func):
        """Returns the full path of the function as a string."""
        module = inspect.getmodule(func).__name__
        func_name = func.__name__  # Use __name__ instead of __qualname__
        return f"{module}.{func_name}"


# Usage Example
# task_instance = Task()
# result = task_instance.queue_task('my_module.my_function', arg1, arg2)  # Using string
# result = task_instance.queue_task(my_function, arg1, arg2)  # Using callable



from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render

from core.api.public import APIAuthToken
from core.api.public.permissions import SCOPE_DESCRIPTIONS
from core.models import UserSettings, Organization
from core.service.modals.registry import Modal
from core.types.requests import WebRequest


# class LogoutModal(Modal):
#     modal_name = 'logout'
#
#     def get(self, request: WebRequest, *args, **kwargs):
#         return HttpResponse("logout_modal.html")


class PermissionModalContext:
    def get_context(self, request: WebRequest) -> dict:
        # example
        # "clients": {
        #     "description": "Access customer details",
        #     "options": ["read", "write"]
        # }
        return {
            "permissions": [
                {"name": group, "description": perms["description"], "options": perms["options"]}
                for group, perms in SCOPE_DESCRIPTIONS.items()
            ],
            "APIAuthToken_types": APIAuthToken.AdministratorServiceTypes,
        }


class GenerateAPIKeyModal(Modal, PermissionModalContext):
    modal_name = "generate_api_key"


class PassTeamIdContext:
    def get_context(self, request: WebRequest) -> dict:
        return {
            "team_id": request.GET.get("team"),
        }


class TeamCreateUserModal(Modal, PermissionModalContext, PassTeamIdContext):
    modal_name = "team_create_user"

    def get(self, request: WebRequest):
        context = self.get_context(request)
        context["team_id"] = request.GET.get("team")
        return self.Response(request, context)


class CreateTeamModal(Modal):
    modal_name = "create_team"


class EditTeamMemberPermissions(Modal, PermissionModalContext):
    modal_name = "edit_team_member_permissions"
    template_name = "modals/edit_team_member_permissions.html"

    def get(self, request: WebRequest):
        context = self.get_context(request)

        team = request.user.logged_in_as_team

        if not team:
            messages.error(request, "You are not logged in as a team")
            return render(request, "core/base/toast.html", context)

        for_user = team.members.filter(id=request.GET.get("user")).first()
        for_user_perms = team.permissions.filter(user=for_user).first()

        if not for_user:
            messages.error(request, "User not found")
            return render(request, "core/base/toast.html", context)

        context["editing_user"] = for_user
        context["user_current_scopes"] = for_user_perms.scopes if for_user_perms else []

        return self.Response(request, context)


class ChangeProfilePictureModal(Modal):
    modal_name = "change_profile_picture"

    def get(self, request: WebRequest):
        context = self.get_context(request)

        try:
            context["users_profile_picture"] = request.user.user_profile.profile_picture_url
        except UserSettings.DoesNotExist:
            pass

        return self.Response(request, context)


class LeaveTeamModal(Modal):
    modal_name = "leave_team"

    def get(self, request: WebRequest):
        context = self.get_context(request)

        if request.user.teams_joined.filter(id=request.GET.get("team")).exists():
            context["team"] = Organization.objects.filter(id=request.GET.get("team")).first()

        return self.Response(request, context)


class InvoiceUserToTeamModal(Modal, PassTeamIdContext):
    modal_name = "invite_user_to_team"



from django.core.management.base import BaseCommand
from django.urls import get_resolver


class Command(BaseCommand):
    help = "List all URL patterns"

    def handle(self, *args, **kwargs):
        url_patterns = get_resolver().url_patterns
        self.stdout.write("List of URL patterns:")
        for pattern in url_patterns:
            self.stdout.write(str(pattern))



from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal
from typing import Literal


@dataclass
class DefaultFeature:
    """
    Single Product, e.g. "invoices-created"
    This is a stripe "PRICE" (part of a PRODUCT)
    """

    slug: str  # Consistent slug across plans
    description: str
    max_limit_per_month: int
    subscription_plan: DefaultSubscriptionPlan


@dataclass
class DefaultFeatureGroup:
    """
    Group of products, e.g. "Invoices"
    """

    name: str
    items: list[DefaultFeature]


@dataclass
class DefaultSubscriptionPlan:
    """
    This is a Stripe PRODUCT
    """

    name: str
    price_per_month: int
    description: str


# Default subscription plans
free_plan = DefaultSubscriptionPlan(
    name="Free Trial",
    price_per_month=0,
    description="Try out MyFinances",
)

starter_plan = DefaultSubscriptionPlan(
    name="Starter",
    price_per_month=5,
    description="For small businesses that need limited features",
)

growth_plan = DefaultSubscriptionPlan(
    name="Growth",
    price_per_month=10,
    description="For growing businesses that need a little extra",
)

# enterprise_plan = DefaultSubscriptionPlan(
#     name="Enterprise",
#     price_per_month=-1,
#     description="Additional customization for your ideal business",
# )

default_subscription_plans: list[DefaultSubscriptionPlan] = [free_plan, starter_plan, growth_plan]

# Default usage plans
default_usage_plans: list[DefaultFeatureGroup] = [
    DefaultFeatureGroup(
        "invoices",
        [
            # region "invoices-created"
            DefaultFeature(
                slug="invoices-created",
                description="Amount of invoices created per month",
                max_limit_per_month=10,
                subscription_plan=free_plan,
            ),
            DefaultFeature(
                slug="invoices-created",
                description="Amount of invoices created per month (starter plan)",
                max_limit_per_month=500,
                subscription_plan=starter_plan,
            ),
            DefaultFeature(
                slug="invoices-created",
                description="Amount of invoices created per month",
                max_limit_per_month=-1,
                subscription_plan=growth_plan,
            ),
            # endregion "invoices-created"
            # region "invoices-sent-via-schedule"
            DefaultFeature(
                slug="invoices-sent-via-schedule",
                description="Amount of invoices sent from a schedule per month",
                max_limit_per_month=1,
                subscription_plan=free_plan,
            ),
            DefaultFeature(
                slug="invoices-sent-via-schedule",
                description="Amount of invoices sent from a schedule per month",
                max_limit_per_month=50,
                subscription_plan=starter_plan,
            ),
            DefaultFeature(
                slug="invoices-sent-via-schedule",
                description="Amount of invoices sent from a schedule per month",
                max_limit_per_month=-1,
                subscription_plan=growth_plan,
            ),
            # endregion "invoices-sent-via-schedule"
        ],
    ),
    DefaultFeatureGroup(
        "teams",
        [
            # region "organization-access"
            DefaultFeature(
                slug="organization-access",
                description="Amount of invoices created per month (starter plan)",
                max_limit_per_month=1,
                subscription_plan=starter_plan,
            ),
            DefaultFeature(
                slug="organization-access",
                description="Amount of invoices created per month",
                max_limit_per_month=1,
                subscription_plan=growth_plan,
            ),
            # endregion "organization-access"
        ],
    ),
]



from dataclasses import dataclass, field
from typing import TypedDict

from mypy_boto3_sesv2.type_defs import SendEmailResponseTypeDef, SendBulkEmailResponseTypeDef

from core.utils.dataclasses import BaseServiceResponse


class SingleEmailSendServiceResponse(BaseServiceResponse[SendEmailResponseTypeDef]): ...


class BulkEmailSendServiceResponse(BaseServiceResponse[SendBulkEmailResponseTypeDef]): ...


class SingleTemplatedEmailContent(TypedDict):
    template_name: str
    template_data: dict | str


@dataclass(frozen=False)
class SingleEmailInput:
    destination: str | list[str]
    subject: str | None
    content: str | SingleTemplatedEmailContent
    ConfigurationSetName: str | None = None
    from_address: str | None = None
    from_address_name_prefix: str | None = None


@dataclass
class BulkEmailEmailItem:
    destination: str
    template_data: dict | str
    cc: list[str] = field(default_factory=list)
    bcc: list[str] = field(default_factory=list)


@dataclass(frozen=False)
class BulkTemplatedEmailInput:
    email_list: list[BulkEmailEmailItem]
    template_name: str
    default_template_data: dict | str
    ConfigurationSetName: str | None = None
    from_address: str | None = None
    from_address_name_prefix: str | None = None



from __future__ import annotations

import itertools
import typing
from datetime import datetime, timedelta
from typing import Literal, Union
from uuid import uuid4

from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import AbstractUser, UserManager
from django.core.files.storage import storages, FileSystemStorage
from django.db import models
from django.db.models import Count, QuerySet
from django.utils import timezone
from django.utils.crypto import get_random_string
from storages.backends.s3 import S3Storage


def _public_storage():
    return storages["public_media"]


def _private_storage() -> FileSystemStorage | S3Storage:
    return storages["private_media"]


def RandomCode(length=6):
    return get_random_string(length=length).upper()


def RandomAPICode(length=89):
    return get_random_string(length=length).lower()


def upload_to_user_separate_folder(instance, filename, optional_actor=None) -> str:
    instance_name = instance._meta.verbose_name.replace(" ", "-")

    print(instance, filename)

    if optional_actor:
        if isinstance(optional_actor, User):
            return f"{instance_name}/users/{optional_actor.id}/{filename}"
        elif isinstance(optional_actor, Organization):
            return f"{instance_name}/orgs/{optional_actor.id}/{filename}"
        return f"{instance_name}/global/{filename}"

    if hasattr(instance, "user") and hasattr(instance.user, "id"):
        return f"{instance_name}/users/{instance.user.id}/{filename}"
    elif hasattr(instance, "organization") and hasattr(instance.organization, "id"):
        return f"{instance_name}/orgs/{instance.organization.id}/{filename}"
    return f"{instance_name}/global/{filename}"


def USER_OR_ORGANIZATION_CONSTRAINT():
    return models.CheckConstraint(
        name=f"%(app_label)s_%(class)s_check_user_or_organization",
        check=(models.Q(user__isnull=True, organization__isnull=False) | models.Q(user__isnull=False, organization__isnull=True)),
    )


def add_3hrs_from_now():
    return timezone.now() + timezone.timedelta(hours=3)


M = typing.TypeVar("M", bound=models.Model)


class CustomUserManager(UserManager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .select_related("user_profile", "logged_in_as_team")
            .annotate(notification_count=(Count("user_notifications")))
        )


class User(AbstractUser):
    objects: CustomUserManager = CustomUserManager()  # type: ignore

    logged_in_as_team = models.ForeignKey("Organization", on_delete=models.SET_NULL, null=True, blank=True)
    stripe_customer_id = models.CharField(max_length=255, null=True, blank=True)
    entitlements = models.JSONField(null=True, blank=True, default=list)  # list of strings e.g. ["invoices"]
    awaiting_email_verification = models.BooleanField(default=True)
    require_change_password = models.BooleanField(default=False)  # does user need to change password upon next login

    class Role(models.TextChoices):
        #        NAME     DJANGO ADMIN NAME
        DEV = "DEV", "Developer"
        STAFF = "STAFF", "Staff"
        USER = "USER", "User"
        TESTER = "TESTER", "Tester"

    role = models.CharField(max_length=10, choices=Role.choices, default=Role.USER)

    @property
    def name(self):
        return self.first_name

    @property
    def teams_apart_of(self):
        return set(itertools.chain(self.teams_joined.all(), self.teams_leader_of.all()))

    @property
    def is_org(self):
        return False


class ActiveManager(models.Manager):
    """Manager to return only active objects."""

    def get_queryset(self):
        return super().get_queryset().filter(active=True)


class ExpiredManager(models.Manager):
    """Manager to return only expired (inactive) objects."""

    def get_queryset(self):
        now = timezone.now()
        return super().get_queryset().filter(expires__isnull=False, expires__lte=now)


class ExpiresBase(models.Model):
    """Base model for handling expiration logic."""

    expires = models.DateTimeField("Expires", null=True, blank=True, help_text="When the item will expire")
    active = models.BooleanField(default=True)

    # Default manager that returns only active items
    objects = ActiveManager()

    # Custom manager to get expired/inactive objects
    expired_objects = ExpiredManager()

    # Fallback All objects
    all_objects = models.Manager()

    def deactivate(self) -> None:
        """Manually deactivate the object."""
        self.active = False
        self.save()

    def delete_if_expired_for(self, days: int = 14) -> bool:
        """Delete the object if it has been expired for a certain number of days."""
        if self.expires and self.expires <= timezone.now() - timedelta(days=days):
            self.delete()
            return True
        return False

    @property
    def remaining_active_time(self):
        """Return the remaining time until expiration, or None if already expired or no expiration set."""
        if not self.has_expired:
            return self.expires - timezone.now()
        return None

    @property
    def has_expired(self):
        return self.expires and self.expires <= timezone.now()

    def is_active(self):
        return self.active

    class Meta:
        abstract = True


class VerificationCodes(ExpiresBase):
    class ServiceTypes(models.TextChoices):
        CREATE_ACCOUNT = "create_account", "Create Account"
        RESET_PASSWORD = "reset_password", "Reset Password"

    uuid = models.UUIDField(default=uuid4, editable=False, unique=True)  # This is the public identifier
    token = models.TextField(default=RandomCode, editable=False)  # This is the private token (should be hashed)

    user = models.ForeignKey(User, on_delete=models.CASCADE)
    created = models.DateTimeField(auto_now_add=True)
    service = models.CharField(max_length=14, choices=ServiceTypes.choices)

    def __str__(self):
        return self.user.username

    def hash_token(self):
        self.token = make_password(self.token)
        self.save()
        return True

    class Meta:
        verbose_name = "Verification Code"
        verbose_name_plural = "Verification Codes"


class UserSettings(models.Model):
    class CoreFeatures(models.TextChoices):
        INVOICES = "invoices", "Invoices"
        RECEIPTS = "receipts", "Receipts"
        EMAIL_SENDING = "email_sending", "Email Sending"
        MONTHLY_REPORTS = "monthly_reports", "Monthly Reports"

    CURRENCIES = {
        "GBP": {"name": "British Pound Sterling", "symbol": "£"},
        "EUR": {"name": "Euro", "symbol": "€"},
        "USD": {"name": "United States Dollar", "symbol": "$"},
        "JPY": {"name": "Japanese Yen", "symbol": "¥"},
        "INR": {"name": "Indian Rupee", "symbol": "₹"},
        "AUD": {"name": "Australian Dollar", "symbol": "$"},
        "CAD": {"name": "Canadian Dollar", "symbol": "$"},
    }
    user = models.OneToOneField(User, on_delete=models.CASCADE, related_name="user_profile")
    dark_mode = models.BooleanField(default=True)
    currency = models.CharField(
        max_length=3,
        default="GBP",
        choices=[(code, info["name"]) for code, info in CURRENCIES.items()],
    )
    profile_picture = models.ImageField(
        upload_to="profile_pictures/",
        storage=_public_storage,
        blank=True,
        null=True,
    )

    disabled_features = models.JSONField(default=list)

    @property
    def profile_picture_url(self):
        if self.profile_picture and hasattr(self.profile_picture, "url"):
            return self.profile_picture.url
        return ""

    def get_currency_symbol(self):
        return self.CURRENCIES.get(self.currency, {}).get("symbol", "$")

    def has_feature(self, feature: str) -> bool:
        return feature not in self.disabled_features

    def __str__(self):
        return self.user.username

    class Meta:
        verbose_name = "User Settings"
        verbose_name_plural = "User Settings"


class Organization(models.Model):
    name = models.CharField(max_length=100, unique=True)
    leader = models.ForeignKey(User, on_delete=models.CASCADE, related_name="teams_leader_of")
    members = models.ManyToManyField(User, related_name="teams_joined")

    stripe_customer_id = models.CharField(max_length=255, null=True, blank=True)
    entitlements = models.JSONField(null=True, blank=True, default=list)  # list of strings e.g. ["invoices"]

    def is_owner(self, user: User) -> bool:
        return self.leader == user

    def is_logged_in_as_team(self, request) -> bool:
        if isinstance(request.auth, User):
            return False

        if request.auth and request.auth.organization_id == self.id:
            return True
        return False

    @property
    def is_authenticated(self):
        return True

    @property
    def is_org(self):
        return True


class TeamMemberPermission(models.Model):
    team = models.ForeignKey("core.Organization", on_delete=models.CASCADE, related_name="permissions")
    user = models.OneToOneField("core.User", on_delete=models.CASCADE, related_name="team_permissions")
    scopes = models.JSONField("Scopes", default=list, help_text="List of permitted scopes")

    class Meta:
        unique_together = ("team", "user")


class TeamInvitation(ExpiresBase):
    code = models.CharField(max_length=10)
    team = models.ForeignKey("core.Organization", on_delete=models.CASCADE, related_name="team_invitations")
    user = models.ForeignKey("core.User", on_delete=models.CASCADE, related_name="team_invitations")
    invited_by = models.ForeignKey("core.User", on_delete=models.CASCADE)

    def is_active(self):
        return self.active

    def set_expires(self):
        self.expires = timezone.now() + timezone.timedelta(days=14)

    def save(self, *args, **kwargs):
        if not self.code:
            self.code = RandomCode(10)
            self.set_expires()
        super().save()

    def __str__(self):
        return self.team.name

    class Meta:
        verbose_name = "Team Invitation"
        verbose_name_plural = "Team Invitations"


class OwnerBaseManager(models.Manager):
    def create(self, **kwargs):
        # Handle the 'owner' argument dynamically in `create()`
        owner = kwargs.pop("owner", None)
        if isinstance(owner, User):
            kwargs["user"] = owner
            kwargs["organization"] = None
        elif isinstance(owner, Organization):
            kwargs["organization"] = owner
            kwargs["user"] = None
        return super().create(**kwargs)

    def filter(self, *args, **kwargs):
        # Handle the 'owner' argument dynamically in `filter()`
        owner = kwargs.pop("owner", None)
        if isinstance(owner, User):
            kwargs["user"] = owner
        elif isinstance(owner, Organization):
            kwargs["organization"] = owner
        return super().filter(*args, **kwargs)


class OwnerBase(models.Model):
    user = models.ForeignKey("core.User", on_delete=models.CASCADE, null=True, blank=True)
    organization = models.ForeignKey("core.Organization", on_delete=models.CASCADE, null=True, blank=True)

    objects = OwnerBaseManager()

    class Meta:
        abstract = True
        constraints = [
            USER_OR_ORGANIZATION_CONSTRAINT(),
        ]

    @property
    def owner(self) -> User | Organization:
        """
        Property to dynamically get the owner (either User or Team)
        """
        if hasattr(self, "user") and self.user:
            return self.user
        elif hasattr(self, "team") and self.team:
            return self.team
        return self.organization  # type: ignore[return-value]
        # all responses WILL have either a user or org so this will handle all

    @owner.setter
    def owner(self, value: User | Organization) -> None:
        if isinstance(value, User):
            self.user = value
            self.organization = None
        elif isinstance(value, Organization):
            self.user = None
            self.organization = value
        else:
            raise ValueError("Owner must be either a User or a Organization")

    def save(self, *args, **kwargs):
        if hasattr(self, "owner") and not self.user and not self.organization:
            if isinstance(self.owner, User):
                self.user = self.owner
            elif isinstance(self.owner, Organization):
                self.organization = self.owner
        super().save(*args, **kwargs)

    @classmethod
    def filter_by_owner(cls: typing.Type[M], owner: Union[User, Organization]) -> QuerySet[M]:
        """
        Class method to filter objects by owner (either User or Organization)
        """
        if isinstance(owner, User):
            return cls.objects.filter(user=owner)  # type: ignore[attr-defined]
        elif isinstance(owner, Organization):
            return cls.objects.filter(organization=owner)  # type: ignore[attr-defined]
        else:
            raise ValueError("Owner must be either a User or an Organization")

    @property
    def is_team(self):
        return isinstance(self.owner, Organization)


class PasswordSecret(ExpiresBase):
    user = models.OneToOneField(User, on_delete=models.CASCADE, related_name="password_secrets")
    secret = models.TextField(max_length=300)


class Client(OwnerBase):
    active = models.BooleanField(default=True)
    name = models.CharField(max_length=64)
    phone_number = models.CharField(max_length=100, blank=True, null=True)
    email = models.EmailField(blank=True, null=True)
    email_verified = models.BooleanField(default=False)
    company = models.CharField(max_length=100, blank=True, null=True)
    contact_method = models.CharField(max_length=100, blank=True, null=True)
    is_representative = models.BooleanField(default=False)

    address = models.TextField(max_length=100, blank=True, null=True)
    city = models.CharField(max_length=100, blank=True, null=True)
    country = models.CharField(max_length=100, blank=True, null=True)

    def __str__(self):
        return self.name

    def has_access(self, user: User) -> bool:
        if not user.is_authenticated:
            return False

        if user.logged_in_as_team:
            return self.organization == user.logged_in_as_team
        else:
            return self.user == user


class Notification(models.Model):
    action_choices = [
        ("normal", "Normal"),
        ("modal", "Modal"),
        ("redirect", "Redirect"),
    ]

    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name="user_notifications")
    message = models.CharField(max_length=100)
    action = models.CharField(max_length=10, choices=action_choices, default="normal")
    action_value = models.CharField(max_length=100, null=True, blank=True)
    extra_type = models.CharField(max_length=100, null=True, blank=True)
    extra_value = models.CharField(max_length=100, null=True, blank=True)
    date = models.DateTimeField(auto_now_add=True)


class AuditLog(OwnerBase):
    action = models.CharField(max_length=300)
    date = models.DateTimeField(auto_now_add=True)

    class Meta:
        constraints: list = []

    def __str__(self):
        return f"{self.action} - {self.date}"


class LoginLog(models.Model):
    class ServiceTypes(models.TextChoices):
        MANUAL = "manual"
        MAGIC_LINK = "magic_link"

    user = models.ForeignKey(User, on_delete=models.CASCADE)
    service = models.CharField(max_length=14, choices=ServiceTypes.choices, default="manual")
    date = models.DateTimeField(auto_now_add=True)


class Error(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    error = models.CharField(max_length=250, null=True)
    error_code = models.CharField(max_length=100, null=True)
    error_colour = models.CharField(max_length=25, default="danger")
    date = models.DateTimeField(auto_now=True)

    def __str__(self):
        return str(self.user_id)


class TracebackError(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE, null=True, blank=True)
    error = models.CharField(max_length=5000, null=True)
    date = models.DateTimeField(auto_now=True)

    def __str__(self):
        return str(self.error)


class FeatureFlags(models.Model):
    name = models.CharField(max_length=100, editable=False, unique=True)
    description = models.TextField(max_length=500, null=True, blank=True, editable=False)
    value = models.BooleanField(default=False)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        verbose_name = "Feature Flag"
        verbose_name_plural = "Feature Flags"

    def __str__(self):
        return self.name

    def enable(self):
        self.value = True
        self.save()

    def disable(self):
        self.value = False
        self.save()


class EmailSendStatus(OwnerBase):
    STATUS_CHOICES = [
        (status, status.title())
        for status in [
            "send",
            "reject",
            "bounce",
            "complaint",
            "delivery",
            "open",
            "click",
            "rendering_failure",
            "delivery_delay",
            "subscription",
            "failed_to_send",
            "pending",
        ]
    ]

    sent_by = models.ForeignKey(User, on_delete=models.SET_NULL, null=True, blank=True, related_name="emails_sent")

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    updated_status_at = models.DateTimeField(auto_now_add=True)

    recipient = models.TextField()
    aws_message_id = models.CharField(max_length=100, null=True, blank=True, editable=False)
    status = models.CharField(max_length=20, choices=STATUS_CHOICES)

    class Meta:
        constraints = [USER_OR_ORGANIZATION_CONSTRAINT()]



from rest_framework.response import Response


def APIResponse(success: bool = True, data: str | dict | None = None, meta=None, status: int = 0, **kwargs) -> Response:
    """

    Returns a rest_framework Response object, but prefills meta (success etc) aswell as the data with KWARGS.

    """
    meta = meta or {}
    if not status and success:
        status = 201
    elif not status:
        status = 400

    if success:
        return Response({"meta": {"success": True, **meta}, "data": {**kwargs} | data if isinstance(data, dict) else {}}, status=status)
    else:
        return Response({"meta": {"success": False}, "error": data}, status=status)



# import os
#
# import stripe
# from django.urls import reverse
#
# from core.models import User
#
# user: User = User.objects.first()
#
# # stripe.billing.MeterEvent.create(
# #     event_name="invoices_created",
# #     payload={"invoices": "250", "stripe_customer_id": user.stripe_customer_id},
# #     # identifier="id"
# # )
# #
# # stripe.billing.Meter.list_event_summaries(
# #     ""
# # )
#
# # a = stripe.Customer.create(
# #     name=user.get_full_name(),
# #     email=user.email
# # )
# #
# # user.stripe_customer_id = a.id
# # user.save()
#
# # print(a)
#
# # stripe.checkout.Session.create(
# #     success_url=os.environ.get("SITE_URL", default="http://127.0.0.1:8000") + reverse("api:public:webhooks:receive_global"),
# #     line_items=[
# #         {
# #             "price": "price_",
# #             "quantity": 1
# #         }
# #     ]
# # )



from django import template

register = template.Library()


@register.filter
def ordinal(value):
    try:
        value = int(value)
        if 10 <= value % 100 <= 20:
            suffix = "th"
        else:
            suffix = {1: "st", 2: "nd", 3: "rd"}.get(value % 10, "th")
        return f"{value}{suffix}"
    except (ValueError, TypeError):
        return value



from django.contrib import messages
from django.shortcuts import redirect

from core.types.requests import WebRequest


def stripe_failed_return_endpoint(request: WebRequest):
    messages.warning(request, "FAILED RESPONSE")
    return redirect("billing:dashboard")



from __future__ import annotations

from django.urls import include
from django.urls import path

urlpatterns = [
    path("base/", include("core.api.base.urls")),
    path("teams/", include("core.api.teams.urls")),
    path("settings/", include("core.api.settings.urls")),
    path("quotas/", include("core.api.quotas.urls")),
    # path("clients/", include("backend_utils.clients.api.urls")),
    path("emails/", include("core.api.emails.urls")),
    path("maintenance/", include("core.api.maintenance.urls")),
    path("landing_page/", include("core.api.landing_page.urls")),
    path("public/", include("core.api.public.urls")),
    #     path("", include("backend_utils.finance.api.urls")),
]

app_name = "api"



import os
from django.core.management.base import BaseCommand
from django.core.management import call_command


class Command(BaseCommand):
    help = "Runs verification tests for view files in backend/tests/views."

    def add_arguments(self, parser):
        parser.add_argument(
            "test_label",
            nargs="?",
            type=str,
            help="Test label for a specific view file.",
        )

    def handle(self, *args, **options):
        test_dir = "backend/tests/views"
        test_label = options["test_label"]

        if test_label:
            self.run_test("backend.tests.views." + test_label)
        else:
            for root, dirs, files in os.walk(test_dir):
                for file in files:
                    if file.endswith(".py") and file != "__init__.py":
                        test_module = file.replace(".py", "")
                        test_label = f"MAIN.tests.views.{test_module}"
                        self.run_test(test_label)

    def run_test(self, test_label):
        self.stdout.write(self.style.SUCCESS(f"Running tests for {test_label}"))
        call_command("test", test_label)



from django.contrib import messages
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse

from core.models import QuotaLimit


def quota_usage_check_under(
    request, limit: str | QuotaLimit, extra_data: str | int | None = None, api=False, htmx=False, add: int = 0
) -> bool | HttpResponse | HttpResponseRedirect:
    try:
        quota_limit = QuotaLimit.objects.get(slug=limit) if isinstance(limit, str) else limit
    except QuotaLimit.DoesNotExist:
        return True

    if not quota_limit.strict_goes_above_limit(request.user, extra=extra_data, add=add):
        return True

    if api and htmx:
        messages.error(request, f"You have reached the quota limit for this service '{quota_limit.name}'")
        return render(request, "base/toast.html", {"autohide": False})
    elif api:
        return HttpResponse(status=403, content=f"You have reached the quota limit for this service '{quota_limit.name}'")
    messages.error(request, f"You have reached the quota limit for this service '{quota_limit.name}'")
    try:
        last_visited_url = request.session["last_visited"]
        current_url = request.build_absolute_uri()
        if last_visited_url != current_url:
            return HttpResponseRedirect(last_visited_url)
    except KeyError:
        pass
    return HttpResponseRedirect(reverse("dashboard"))


def render_quota_error(request, quota_limit):
    messages.error(request, f"You have reached the quota limit for this service '{quota_limit.slug}'")
    return render(request, "partials/messages_list.html", {"autohide": False})


def render_quota_error_response(quota_limit):
    return HttpResponse(status=403, content=f"You have reached the quota limit for this service '{quota_limit.slug}'")



from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.http import require_http_methods

from core.api.public import APIAuthToken
from core.service.api_keys.delete import delete_api_key
from core.service.api_keys.generate import generate_public_api_key
from core.service.api_keys.get import get_api_key_by_id
from core.service.permissions.scopes import get_permissions_from_request

from core.types.requests import WebRequest
from core.decorators import web_require_scopes


@require_http_methods(["POST"])
@web_require_scopes("api_keys:write")
def generate_api_key_endpoint(request: WebRequest) -> HttpResponse:
    name = request.POST.get("name")
    expires = request.POST.get("expires")
    description = request.POST.get("description")
    administrator_toggle = True if request.POST.get("administrator") == "on" else False
    administrator_type = request.POST.get("administrator_type")

    permissions: list = get_permissions_from_request(request)

    key_obj, key_response = generate_public_api_key(
        request,
        request.user.logged_in_as_team or request.user,
        name,
        permissions,
        expires=expires,
        description=description,
        administrator_toggle=administrator_toggle,
        administrator_type=administrator_type,
    )

    if not key_obj:
        messages.error(request, key_response)
        return render(request, "base/toast.html")

    messages.success(request, "API key generated successfully")

    http_response = render(
        request,
        "pages/settings/settings/api_key_generated_response.html",
        {
            "raw_key": key_response,
            "name": name,
        },
    )

    http_response.headers["HX-Reswap"] = "beforebegin"
    http_response.headers["HX-Retarget"] = 'div[data-hx-container="api_keys"]'

    return http_response


@require_http_methods(["DELETE"])
def revoke_api_key_endpoint(request: WebRequest, key_id: str) -> HttpResponse:
    key: APIAuthToken | None = get_api_key_by_id(request.user.logged_in_as_team or request.user, key_id)

    delete_key_response = delete_api_key(request, request.user.logged_in_as_team or request.user, key=key)

    if isinstance(delete_key_response, str):
        messages.error(request, "This key does not exist")
    else:
        messages.success(request, "Successfully revoked the API Key")
    return render(request, "base/toast.html")



from django.contrib import messages
from django.contrib.auth.hashers import check_password
from django.http import HttpRequest
from django.shortcuts import render, redirect
from django.utils import timezone

from core.models import PasswordSecret
from core.decorators import not_authenticated


@not_authenticated
def set_password(request: HttpRequest, secret):
    SECRET_RETURNED = PasswordSecret.objects.all()
    SECRET_RETURNED.filter(expires__lte=timezone.now()).all().delete()

    for SECRET in SECRET_RETURNED:
        if check_password(secret, SECRET.secret):
            return render(request, "pages/reset_password.html", {"secret": secret})

    messages.error(request, "Invalid or expired password reset code")
    return redirect("dashboard")



from __future__ import annotations

from django.conf import settings
from django.core.cache import cache
from django.core.cache.backends.redis import RedisCacheClient

from core.utils.settings_helper import send_email

cache: RedisCacheClient = cache
from django.core.files.storage import default_storage
from django.db.models.signals import pre_save, post_delete, post_save, pre_delete
from django.dispatch import receiver
from django.urls import reverse

from core.models import UserSettings, User, FeatureFlags, VerificationCodes


@receiver(pre_save, sender=UserSettings)
def delete_old_profile_picture(sender, instance, **kwargs):
    if not instance.pk:
        return

    try:
        old_profile = UserSettings.objects.get(pk=instance.pk)
    except UserSettings.DoesNotExist:
        return

    if old_profile.profile_picture and old_profile.profile_picture != instance.profile_picture:
        # If the profile picture has been updated, delete the old file
        old_profile.profile_picture.delete(save=False)


@receiver(post_delete, sender=UserSettings)
def set_profile_picture_to_none(sender, instance, **kwargs):
    # Check if the file exists in the storage
    if instance.profile_picture and default_storage.exists(instance.profile_picture.name):
        instance.profile_picture.delete(save=False)


@receiver(post_save, sender=User)
def user_account_create_make_usersettings(sender, instance, created, **kwargs):
    if created:
        try:
            users_settings = instance.user_profile
        except UserSettings.DoesNotExist:
            users_settings = None

        if not users_settings:
            UserSettings.objects.create(user=instance)


@receiver(post_save, sender=FeatureFlags)
def refresh_feature_cache(sender, instance: FeatureFlags, **kwargs):
    feature = instance.name
    key = f"myfinances:feature_flag:{feature}"

    cached_value = cache.get(key)

    if cached_value:
        return cache.delete(key)


@receiver(post_save, sender=User)
def send_welcome_email(sender, instance: User, created, **kwargs):
    if created:
        email_message = f"""
            Welcome to MyFinances{f", {instance.first_name}" if instance.first_name else ""}!

            We're happy to have you join us. We are still in development and working on the core features.

            In app we have a live chat, so please drop us a message or email support@myfinances.cloud if you have any queries.

            Thank you for using MyFinances!
        """
        magic_link = VerificationCodes.objects.create(user=instance, service="create_account")
        token_plain = magic_link.token
        magic_link.hash_token()
        magic_link_url = settings.SITE_URL + reverse(
            "core:auth:login create_account verify", kwargs={"uuid": magic_link.uuid, "token": token_plain}
        )
        email_message += f"""
            To start with, you must first **verify this email** so that we can link your account to this email.
            Click the link below to activate your account, no details are required, once pressed you're all set!

            Verify Link: {magic_link_url}
        """

        email = send_email(destination=instance.email, subject="Welcome to MyFinances", content=email_message)

        #     User.send_welcome_email(instance)



from __future__ import annotations

from django.urls import path

from . import dashboard

urlpatterns = [
    path(
        "",
        dashboard.dashboard,
        name="dashboard",
    ),
]

app_name = "emails"



from drf_yasg.utils import swagger_auto_schema
from drf_yasg import openapi
from django.db import connection, OperationalError
from django.core.cache import cache

from rest_framework.decorators import api_view, permission_classes

from core.api.public.permissions import IsSuperuser
from core.api.public.helpers.response import APIResponse
from core.api.public.swagger_ui import API_RESPONSE_BASE


@swagger_auto_schema(
    method="get",
    operation_description="Check the system's health by verifying database and external API connections.",
    responses={
        200: openapi.Response(
            description="System health check result",
            schema=openapi.Schema(
                type=openapi.TYPE_OBJECT,
                properties=API_RESPONSE_BASE(
                    {
                        "problems": openapi.Schema(
                            type=openapi.TYPE_ARRAY,
                            items=openapi.Schema(
                                type=openapi.TYPE_OBJECT,
                                properties={
                                    "id": openapi.Schema(type=openapi.TYPE_STRING, description="Problem ID"),
                                    "message": openapi.Schema(type=openapi.TYPE_STRING, description="Problem message"),
                                },
                            ),
                        ),
                        "healthy": openapi.Schema(type=openapi.TYPE_BOOLEAN, description="Indicates overall system health"),
                    }
                ),
            ),
            examples={
                "application/json": {
                    "problems": [
                        {"id": "database", "message": "database failed to connect"},
                    ],
                    "healthy": False,
                }
            },
        )
    },
)
@api_view(["GET"])
@permission_classes([IsSuperuser])
def system_health_endpoint(request):
    if not request.user or not request.user.is_superuser:
        return APIResponse(False, "User is not permitted to view internal information", status=403)

    problems = []

    try:
        connection.ensure_connection()
    except OperationalError:
        problems.append({"id": "database", "message": "database failed to connect"})

    try:
        cache._cache.get_client().ping()
    except ConnectionError:
        problems.append({"id": "redis", "message": "redis failed to connect"})

    return APIResponse(True, {"problems": problems, "healthy": not bool(problems)})



import importlib

from django.apps import AppConfig
from django.conf import settings


class BillingConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "billing"

    def ready(self):
        from . import signals

        self._load_project_signals()

    def _load_project_signals(self):
        for app_name in getattr(settings, "INSTALLED_APPS", []):
            try:
                module_name = f"{app_name}.billing.signals"
                importlib.import_module(module_name)
            except ModuleNotFoundError:
                pass



from __future__ import annotations

from django.conf import settings
from django.conf.urls.static import static
from django.urls import include
from django.urls import path
from django.urls import re_path as url
from django.views.generic import RedirectView
from django.views.static import serve

from core.api.public.swagger_ui import get_swagger_ui, get_swagger_endpoints

url(
    r"^frontend/static/(?P<path>.*)$",
    serve,
    {"document_root": settings.STATICFILES_DIRS[0]},
)
urlpatterns = [
    # path("tz_detect/", include("tz_detect.urls")),
    # path("pricing", pricing, name="pricing"),
    path("dashboard/settings/", include("core.views.settings.urls")),
    path("dashboard/teams/", include("core.views.teams.urls")),
    path("dashboard/emails/", include("core.views.emails.urls")),
    path("favicon.ico", RedirectView.as_view(url=settings.STATIC_URL + "favicon.ico")),
    path("login/external/", include("social_django.urls", namespace="social")),
    path("auth/", include("core.views.auth.urls")),
    path("api/", include("core.api.urls")),
] + static(settings.STATIC_URL, document_root=settings.STATICFILES_DIRS[0])

app_name = "core"

if settings.DEBUG:
    urlpatterns += [path("silk/", include("silk.urls", namespace="silk"))]

    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)
    # may not need to be in debug
    urlpatterns += static(settings.STATIC_URL, document_root=settings.STATICFILES_DIRS[0])

schema_view = get_swagger_ui()
urlpatterns += get_swagger_endpoints(settings.DEBUG)

handler500 = "core.views.other.errors.universal"
handler404 = "core.views.other.errors.universal"
handler403 = "core.views.other.errors.e_403"



from typing import Callable, TypeVar
from core.utils.dataclasses import BaseServiceResponse

T = TypeVar("T", bound=BaseServiceResponse)


def retry_handler(function: Callable[..., T], *args, retry_max_attempts: int = 3, **kwargs) -> T:
    attempts: int = 0

    while attempts < retry_max_attempts:
        response: T = function(*args, **kwargs)

        if response.failed:
            attempts += 1
            if attempts == retry_max_attempts:
                return response
            continue
        return response
    return response



from django.db.models import QuerySet

from core.api.public import APIAuthToken
from core.models import UserSettings

# from core.service.defaults.get import get_account_defaults
from core.types.requests import WebRequest


def validate_page(page: str | None) -> bool:
    return not page or page in ["profile", "account", "api_keys", "account_defaults", "account_security", "email_templates"]


def get_user_profile(request: WebRequest) -> UserSettings:
    try:
        usersettings = request.user.user_profile
    except UserSettings.DoesNotExist:
        # Create a new UserSettings object
        usersettings = UserSettings.objects.create(user=request.user)
    return usersettings


def get_api_keys(request: WebRequest) -> QuerySet[APIAuthToken]:
    return APIAuthToken.filter_by_owner(request.actor).filter(active=True).only("created", "name", "last_used", "description", "expires")


def get_account_page_context(request: WebRequest, context: dict) -> None:
    user_profile = get_user_profile(request)
    context.update({"currency_signs": user_profile.CURRENCIES, "currency": user_profile.currency})


def api_keys_page_context(request: WebRequest, context: dict) -> None:
    api_keys = get_api_keys(request)
    context.update({"api_keys": api_keys})


def account_defaults_context(request: WebRequest, context: dict) -> None:
    ...
    # context.update({"account_defaults": get_account_defaults(request.actor)})


def profile_context(request: WebRequest, context: dict) -> None: ...


def account_security_context(request: WebRequest, context: dict) -> None: ...



INSTALLED_APPS = ["django.contrib.auth", "django.contrib.contenttypes", "django.contrib.staticfiles", "core", "billing"]

STORAGES = {
    "default": {
        "BACKEND": "django.core.files.storage.FileSystemStorage",
    },
    "staticfiles": {
        "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
    },
    "public_media": {
        "BACKEND": "django.core.files.storage.FileSystemStorage",
        "OPTIONS": {"location": "media/public", "base_url": "/media/public/"},
    },
    "private_media": {
        "BACKEND": "django.core.files.storage.FileSystemStorage",
        "OPTIONS": {"location": "media/private", "base_url": "/media/private/"},
    },
}

AUTH_USER_MODEL = "core.User"

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"



from django.urls import path

from core.api.maintenance.now import handle_maintenance_now_endpoint

urlpatterns = [
    path("cleanup/", handle_maintenance_now_endpoint, name="cleanup"),
]

app_name = "maintenance"



from rest_framework.request import Request

from core.api.public import APIAuthToken
from core.models import User, Organization


class APIRequest(Request):
    user: User
    auth: APIAuthToken
    api_token: APIAuthToken
    team: Organization | None
    team_id: int | None



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class FeatureFlag:
    name: str
    description: str
    default: bool


default_feature_flags: list[FeatureFlag] = [
    FeatureFlag(name="areSignupsEnabled", description="Are new account creations allowed", default=True),
    FeatureFlag(
        name="isInvoiceSchedulingEnabled",
        description="Invoice Scheduling allows for clients to create invoice schedules that send and invoice at a specific date.",
        default=False,
    ),
    FeatureFlag(name="areUserEmailsAllowed", description="Are users allowed to send emails from YOUR DOMAIN to customers", default=False),
    FeatureFlag(
        name="areInvoiceRemindersEnabled",
        description="Invoice Reminders allow for clients to be reminded to pay an invoice.",
        default=False,
    ),
]



from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render

from core.models import Notification
from core.types.htmx import HtmxHttpRequest


def get_notification_html(request: HtmxHttpRequest):
    user_notifications = Notification.objects.filter(user=request.user).order_by("-date")
    count = user_notifications.count()

    if count > 5:
        user_notifications = user_notifications[:5]

    return render(
        request,
        "core/base/topbar/_notification_dropdown_items.html",
        {"notifications": user_notifications, "notif_count": count},
    )


def get_notification_count_html(request: HtmxHttpRequest):
    user_notifications = Notification.objects.filter(user=request.user).count()
    return HttpResponse(f"{user_notifications}")


def delete_notification(request: HtmxHttpRequest, id: int):
    notif = Notification.objects.filter(id=id, user=request.user).first()

    if notif is None or notif.user != request.user:
        if request.htmx:
            messages.error(request, "Notification not found")
            return render(request, "base/toasts.html")
        return HttpResponse(status=404, content="Notification not found")

    notif.delete()

    response = HttpResponse(status=200)
    response["HX-Trigger"] = "refresh_notification_count"
    return response



from django.urls import path

from . import login, create_account, verify
from .passwords import view as passwords_view, generate as passwords_generate, set as passwords_set

urlpatterns = [
    path("login/", login.login_initial_page, name="login"),
    path("login/manual/", login.login_manual, name="login manual"),
    path("login/magic_link/request/", login.MagicLinkRequestView.as_view(), name="login magic_link request"),
    path("login/magic_link/request/wait/", login.MagicLinkWaitingView.as_view(), name="login magic_link request wait"),
    path("login/magic_link/verify/<uuid:uuid>/<str:token>/", login.MagicLinkVerifyView.as_view(), name="login magic_link verify"),
    path(
        "login/magic_link/verify/<uuid:uuid>/<str:token>/accept/",
        login.MagicLinkVerifyAccept.as_view(),
        name="login magic_link verify accept",
    ),
    path(
        "login/magic_link/verify/<uuid:uuid>/<str:token>/decline/",
        login.MagicLinkVerifyDecline.as_view(),
        name="login magic_link verify decline",
    ),
    path(
        "login/forgot_password/",
        login.forgot_password_page,
        name="login forgot_password",
    ),
    path("logout/", login.logout_view, name="logout"),
    path(
        "create_account/",
        create_account.CreateAccountChooseView.as_view(),
        name="login create_account",
    ),
    path(
        "create_account/manual/",
        create_account.CreateAccountManualView.as_view(),
        name="login create_account manual",
    ),
    path(
        "create_account/verify/<uuid:uuid>/<str:token>/",
        verify.create_account_verify,
        name="login create_account verify",
    ),
    path(
        "create_account/verify/resend/",
        verify.resend_verification_code,
        name="login create_account verify resend",
    ),
    # path(
    #     "login/magic_link/<str:uuid>/",
    #     login.magic_link,
    #     name="login magic_link",
    # )
    path(
        "reset-password/",
        passwords_generate.password_reset,
        name="user set password reset",
    ),
    path(
        "set-password/<str:secret>/",
        passwords_view.set_password,
        name="user set password",
    ),
    path(
        "set-password/<str:secret>/set/",
        passwords_set.set_password_set,
        name="user set password set",
    ),
    path(
        "admin/generate-password/",
        passwords_generate.set_password_generate,
        name="admin set password generate",
    ),
]

app_name = "auth"



from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.http import require_http_methods

from core.types.htmx import HtmxHttpRequest


@require_http_methods(["POST"])
def change_account_name(request: HtmxHttpRequest):
    if not request.htmx:
        return HttpResponse("Invalid Request", status=405)

    htmx_return = "base/toasts.html"

    first_name = request.POST.get("first_name")
    last_name = request.POST.get("last_name")

    if not first_name and not last_name:
        messages.error(request, "Please enter a valid firstname or lastname.")
        return render(request, htmx_return)

    if request.user.first_name == first_name and request.user.last_name == last_name:
        messages.warning(request, "You already have this name.")
        return render(request, htmx_return)

    if first_name:
        request.user.first_name = first_name

    if last_name:
        request.user.last_name = last_name

    request.user.save()

    messages.success(
        request,
        f"Successfully changed your name to <strong>{request.user.get_full_name()}</strong>",
    )

    return render(request, htmx_return)



from django.urls import path

from core.views.settings.teams import teams_dashboard_handler

urlpatterns = [
    path(
        "",
        teams_dashboard_handler,
        name="dashboard",
    ),
]

app_name = "teams"



import stripe


def get_price_id_from_lookup_key(lookup_key: str) -> str:
    prices = stripe.Price.list(lookup_keys=[lookup_key])
    if prices.data:
        return prices.data[0].id  # Assuming the lookup key returns one price
    raise ValueError(f"Price with lookup key {lookup_key} not found.")



from datetime import datetime
import stripe
from django.db.models import QuerySet
from core.models import User, Organization
from billing.models import UserSubscription, SubscriptionPlan
from billing.service.entitlements import update_user_entitlements
from billing.service.stripe_customer import get_or_create_customer_id


def handle_plan_change(user_subscription: UserSubscription, new_plan: SubscriptionPlan) -> UserSubscription:
    """
    Handles plan upgrades or downgrades.
    """
    # Cancel the current Stripe subscription if necessary
    stripe.Subscription.modify(
        user_subscription.stripe_subscription_id,
        cancel_at_period_end=False,  # Cancels immediately
    )

    # Create a new Stripe subscription for the new plan
    new_subscription = stripe.Subscription.create(
        customer=user_subscription.owner.stripe_customer_id,
        items=[{"price": new_plan.stripe_price_id}],
    )

    # Update the UserSubscription object with new plan and subscription id
    user_subscription.subscription_plan = new_plan
    user_subscription.stripe_subscription_id = new_subscription.id
    user_subscription.save()

    return user_subscription



from __future__ import annotations

from django.http import HttpResponse
from django.shortcuts import render

from core.decorators import feature_flag_check, web_require_scopes
from core.types.htmx import HtmxHttpRequest


@feature_flag_check("areUserEmailsAllowed", status=True)
@web_require_scopes("emails:read", False, False, "dashboard")
def dashboard(request: HtmxHttpRequest) -> HttpResponse:
    return render(request, "pages/emails/dashboard.html", {})



import stripe.entitlements
from django.core.cache import cache
from django.core.cache.backends.redis import RedisCacheClient

from core.models import User, Organization
from billing.models import StripeWebhookEvent
from billing.service.get_user import get_actor_from_stripe_customer

cache: RedisCacheClient = cache


def entitlements_updated_via_stripe_webhook(webhook_event: StripeWebhookEvent) -> None:
    data: stripe.entitlements.ActiveEntitlementSummary = webhook_event.data["object"]
    actor = get_actor_from_stripe_customer(data["customer"])

    if not actor:
        print("No actor found for customer.")
        return

    # Re-fetch and update the entitlements for the actor (User or Organization)
    update_user_entitlements(actor)


def update_user_entitlements(actor: User | Organization) -> list[str]:
    if not actor.stripe_customer_id:
        return []

    entitlements = stripe.entitlements.ActiveEntitlement.list(customer=actor.stripe_customer_id, limit=25).data

    entitlement_names = [entitlement.lookup_key for entitlement in entitlements]

    actor.entitlements = entitlement_names
    actor.save(update_fields=["entitlements"])

    cache_actor_type = "user" if isinstance(actor, User) else "org"

    cache.set(f"myfinances:entitlements:{cache_actor_type}:{actor.id}", entitlement_names, timeout=3600)

    return entitlement_names


def get_entitlements(actor: User | Organization, avoid_cache=False) -> list[str]:
    cache_key = "user" if isinstance(actor, User) else "org"

    if not avoid_cache and (cached_entitlements := cache.get(f"myfinances:entitlements:{cache_key}:{actor.id}", default=[])):
        return cached_entitlements
    return update_user_entitlements(actor)


def has_entitlement(actor: User | Organization, entitlement: str) -> bool:
    return entitlement in get_entitlements(actor)


def has_entitlements(actor: User | Organization, entitlements: list[str]) -> bool:
    return all(entitlement in entitlements for entitlement in get_entitlements(actor))



from typing import Dict, List

from django.conf import settings


class CoreConfig:
    BILLING_ENABLED: bool = False
    EXPIRY_MODELS: List[str] = [
        "core.TeamInvitation",
        "core.PasswordSecret",
    ]
    SETTINGS_PAGE_CONTEXT_HANDLERS: Dict[str, str] = {
        "account": "core.service.settings.view.get_account_page_context",
        "api_keys": "core.service.settings.view.api_keys_page_context",
        "account_defaults": "core.service.settings.view.account_defaults_context",
        "profile": "core.service.settings.view.profile_context",
        "account_security": "core.service.settings.view.account_security_context",
    }

    # DEFAULT_AUTO_FIELD: str = "django.db.models.BigAutoField"
    # SOCIAL_AUTH_USER_MODEL: str = "core.User"
    # AUTH_USER_MODEL: str = "core.User"
    # LOGGING: Dict[str, Dict] = {
    #     "version": 1,
    #     "disable_existing_loggers": False,
    #     "handlers": {
    #         "console": {
    #             "level": "DEBUG",
    #             "class": "logging.StreamHandler",
    #         },
    #     },
    #     "loggers": {
    #         "django": {
    #             "handlers": ["console"],
    #             "level": "DEBUG",
    #             "propagate": True,
    #         },
    #         "django.request": {
    #             "handlers": ["console"],
    #             "level": "ERROR",
    #             "propagate": False,
    #         },
    #     },
    # }

    def __init__(self):
        self._setup()

    def _setup(self):
        from django.conf import settings

        user_options = {option[5:]: getattr(settings, option) for option in dir(settings) if option.startswith("CORE_")}

        for key, value in user_options.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def __getattr__(self, item: str):
        raise AttributeError(f"'CoreConfig' object has no attribute '{item}'")



from django.contrib import messages
from django.http import HttpRequest
from django.shortcuts import redirect

from core.decorators import web_require_scopes
from core.models import User, Organization


@web_require_scopes("team:kick", True, True)
def kick_user(request: HttpRequest, user_id):
    user: User | None = User.objects.filter(id=user_id).first()
    confirmation_text = request.POST.get("confirmation_text")
    if not user:
        messages.error(request, "User not found")
        return redirect("core:teams:dashboard")

    if confirmation_text != f"i confirm i want to kick {user.username}":
        messages.error(request, "Invalid confirmation")
        return redirect("core:teams:dashboard")

    team: Organization | None = user.teams_joined.first()
    if not team:
        messages.error(request, "User is not apart of your team")
        return redirect("core:teams:dashboard")

    if team.leader != request.user:
        messages.error(request, "You don't have the required permissions to kick this user")
        return redirect("core:teams:dashboard")

    team.members.remove(user)
    messages.success(request, f"Successfully kicked {user.username}")

    return redirect("core:teams:dashboard")



import stripe

from core.utils.settings_helper import get_var

STRIPE_TEST_SECRET_KEY = get_var("STRIPE_TEST_SECRET_KEY")
STRIPE_LIVE_SECRET_KEY = get_var("STRIPE_LIVE_SECRET_KEY")
STRIPE_WEBHOOK_ENDPOINT_SECRET = get_var("STRIPE_WEBHOOK_ENDPOINT_SECRET")

STRIPE_MAIN_API_KEY = STRIPE_LIVE_SECRET_KEY if STRIPE_LIVE_SECRET_KEY else STRIPE_TEST_SECRET_KEY

STRIPE_LIVE_MODE = True if STRIPE_LIVE_SECRET_KEY else False

stripe.api_key = STRIPE_MAIN_API_KEY

NO_SUBSCRIPTION_PLAN_DENY_VIEW_NAMES: set[str] = {
    "clients:create",
    "file_storage:upload:start_batch",
    "file_storage:upload:end_batch",
    "file_storage:upload:add_to_batch",
    "file_storage:upload:dashboard",
    # "finance:invoices:single:manage_access",
    "finance:invoices:single:manage_access create",
    # "finance:invoices:single:manage_access delete",
    "finance:invoices:single:edit",
    "finance:invoices:single:create",
    "finance:invoices:recurring:create",
    "finance:invoices:recurring:edit",
    # APIS
    "teams:invite",
    "teams:create",
    "receipts:edit",
    "receipts:new",
    "finance:invoices:single:edit",
    "finance:invoices:single:edit discount",
    "finance:invoices:recurring:generate next invoice",
    "finance:invoices:recurring:edit",
    "finance:invoices:create:set_destination from",
    "finance:invoices:create:set_destination to",
    "finance:invoices:create:services add",
    "products:create",
    "public:clients:create",
    "public:invoices:create",
}



from django.contrib import messages
from django.shortcuts import render

from core.decorators import web_require_scopes
from core.models import Organization
from core.service.permissions.scopes import get_permissions_from_request
from core.service.teams.create_user import create_user_service
from core.types.requests import WebRequest


@web_require_scopes("team:invite", True, True)
def create_user_endpoint(request: WebRequest):
    team_id = request.POST.get("team_id", "")

    team: Organization | None = Organization.objects.filter(id=team_id).first()

    if not team:
        messages.error(request, "This team does not exist")
        return render(request, "base/toast.html")

    if not team.is_owner(request.user):
        messages.error(request, "Only the team owner can create users")
        return render(request, "base/toast.html")

    first_name = request.POST.get("first_name", "")
    last_name = request.POST.get("last_name", "")
    email = request.POST.get("email", "")
    permissions: list = get_permissions_from_request(request)

    created_user = create_user_service(request, email, team, first_name, last_name, permissions)

    if created_user.failed:
        messages.error(request, created_user.error)
        return render(request, "base/toast.html")
    else:
        messages.success(request, f"The account for {first_name} was created successfully. They have been emailed instructions.")
    return render(request, "base/toast.html")



from textwrap import dedent

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.hashers import check_password
from django.shortcuts import redirect
from django.urls import reverse
from django.views.decorators.http import require_POST
from django_ratelimit.decorators import ratelimit

from core.models import VerificationCodes, User, TracebackError
from core.utils.settings_helper import send_email, ARE_EMAILS_ENABLED


def create_account_verify(request, uuid, token):
    object = VerificationCodes.objects.filter(uuid=uuid, service="create_account").first()

    if not object:
        messages.error(request, "Invalid URL")  # Todo: add some way a user can resend code?
        return redirect("core:auth:login create_account")

    if not object.is_active():
        messages.error(request, "This code has already expired")  # Todo: add some way a user can resend code?
        return redirect("core:auth:login create_account")

    if not object.user.awaiting_email_verification:
        messages.error(request, "Your email has already been verified. You can login.")
        return redirect("core:auth:login")

    if not check_password(token, object.token):
        messages.error(request, "This verification token is invalid.")
        return redirect("core:auth:login create_account")

    user = object.user
    user.is_active = True
    user.awaiting_email_verification = False
    user.save()
    object.delete()

    messages.success(request, "Successfully verified your email! You can now login.")
    return redirect("core:auth:login")


def create_magic_link(user: User, service: str) -> tuple[VerificationCodes, str]:
    magic_link = VerificationCodes.objects.create(user=user, service=service)
    token_plain = magic_link.token
    magic_link.hash_token()
    return magic_link, token_plain


@ratelimit(group="resend_verification_code", key="ip", rate="1/m")
@ratelimit(group="resend_verification_code", key="ip", rate="3/25m")
@ratelimit(group="resend_verification_code", key="ip", rate="10/6h")
@ratelimit(group="resend_verification_code", key="post:email", rate="1/m")
@ratelimit(group="resend_verification_code", key="post:email", rate="3/25m")
@require_POST
def resend_verification_code(request):
    email = request.POST.get("email")
    if not email:
        messages.error(request, "Invalid resend verification request")
        return redirect("core:auth:login")
    if not ARE_EMAILS_ENABLED:
        messages.error(request, "Emails are currently disabled.")
        TracebackError.objects.create(error="Emails are currently disabled.")
        return redirect("core:auth:login create_account")
    try:
        user = User.objects.get(email=email)
    except User.DoesNotExist:
        messages.error(request, "Invalid resend verification request")
        return redirect("core:auth:login create_account")
    VerificationCodes.objects.filter(user=user, service="create_account").delete()
    magic_link = create_magic_link(user, "create_account")
    magic_link_url = settings.SITE_URL + reverse(
        "core:auth:login create_account verify", kwargs={"uuid": magic_link.uuid, "token": token_plain}
    )

    send_email(
        destination=email,
        subject="Verify your email",
        content=dedent(
            f"""
                Hi {user.first_name if user.first_name else "User"},

                Verification for your email has been requested to link this email to your MyFinances account.
                If this wasn't you, you can simply ignore this email.

                If it was you, you can complete the verification by clicking the link below.
                Verify Link: {magic_link_url}
            """
        ),
    )

    messages.success(request, "Verification email sent, check your inbox or spam!")
    return redirect("core:auth:login")



from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from login_required import login_not_required

from core.service.maintenance.tasks import execute_maintenance_tasks, get_maintenance_tasks
from core.service.webhooks.auth import authenticate_api_key

from core.service.maintenance.expire.run import expire_and_cleanup_objects

import logging

from core.types.requests import WebRequest

logger = logging.getLogger(__name__)


@require_POST
@csrf_exempt
@login_not_required
def handle_maintenance_now_endpoint(request: WebRequest):
    logger.info("Received routine cleanup handler. Now authenticating...")
    api_auth_response = authenticate_api_key(request)

    if api_auth_response.failed:
        logger.info(f"Maintenance auth failed: {api_auth_response.error}")
        return JsonResponse({"message": api_auth_response.error, "success": False}, status=api_auth_response.status_code or 400)

    output_str = execute_maintenance_tasks(get_maintenance_tasks())
    logger.info(output_str)
    return JsonResponse({"message": output_str, "success": True}, status=200)



from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render

from core.models import Organization
from core.types.htmx import HtmxHttpRequest


def switch_team(request: HtmxHttpRequest, team_id: str | int | None = None):
    if not team_id:
        team_id = request.POST.get("join_team", None)

    if not team_id:
        if not request.user.logged_in_as_team:
            messages.warning(request, "You are not logged into an organization")
        else:
            messages.success(request, "You are now logged into your personal account")

        request.user.logged_in_as_team = None
        request.user.save()
        response = HttpResponse(status=200)
        response["HX-Refresh"] = "true"
        return response

    team: Organization | None = Organization.objects.filter(id=team_id).first()

    if not team:
        messages.error(request, "Team not found")
        return render(request, "partials/messages_list.html")

    if request.user.logged_in_as_team == team:
        messages.error(request, "You are already logged in for this team")
        return render(request, "partials/messages_list.html")

    if not request.user.teams_leader_of.filter(id=team_id).exists() and not request.user.teams_joined.filter(id=team_id).exists():
        messages.error(request, "You are not a member of this team")
        return render(request, "partials/messages_list.html")

    messages.success(request, f"Now signing into the organization '{team.name}'")
    request.user.logged_in_as_team = team
    request.user.save()

    response = HttpResponse(status=200)
    response["HX-Refresh"] = "true"
    return response
    # return render(request, "components/+logged_in_for.html")


def get_dropdown(request: HtmxHttpRequest):
    if not request.htmx:
        return HttpResponse("Invalid Request", status=405)

    return render(request, "base/topbar/_organizations_list.html")



from typing import Type

from rest_framework.authentication import TokenAuthentication, get_authorization_header
from rest_framework.exceptions import AuthenticationFailed
from django.utils.translation import gettext_lazy as _
from core.api.public.models import APIAuthToken
from core.models import User, Organization

from rest_framework import exceptions


class CustomBearerAuthentication(TokenAuthentication):
    keyword = "Bearer"

    def get_model(self) -> Type[APIAuthToken]:
        return APIAuthToken

    def authenticate(self, request):
        auth = get_authorization_header(request).split()

        if not auth or auth[0].lower() != self.keyword.lower().encode():
            return None

        if len(auth) == 1:
            msg = _("Invalid token header. No credentials provided.")
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = _("Invalid token header. Token string should not contain spaces.")
            raise exceptions.AuthenticationFailed(msg)

        try:
            token = auth[1].decode()
        except UnicodeError:
            msg = _("Invalid token header. Token string should not contain invalid characters.")
            raise exceptions.AuthenticationFailed(msg)

        user_or_org, token = self.authenticate_credentials(token)

        request.actor = user_or_org

        if isinstance(user_or_org, Organization):
            request.team = user_or_org
            request.team_id = user_or_org.id
        else:
            request.team = None
            request.team_id = None

        return (user_or_org, token)

    def authenticate_credentials(self, raw_key) -> tuple[User | Organization | None, APIAuthToken]:
        model = self.get_model()

        try:
            token = model.objects.get(hashed_key=model.hash_raw_key(raw_key), active=True)
        except model.DoesNotExist:
            raise AuthenticationFailed(_("Invalid token."))

        if token.has_expired:
            raise AuthenticationFailed(_("Token has expired."))

        user_or_org = token.user or token.organization

        if user_or_org is None:
            raise AuthenticationFailed(_("Associated user or organization not found."))

        return user_or_org, token



from __future__ import annotations

import logging

from django.db.models.signals import post_migrate
from django.dispatch import receiver

from billing.data.default_usage_plans import default_usage_plans, default_subscription_plans, DefaultFeature, DefaultSubscriptionPlan
from billing.models import PlanFeature, PlanFeatureGroup, SubscriptionPlan


#
@receiver(post_migrate)
def update_usage_plans(**kwargs):
    subscription_plans: dict = {}

    subscription_plan: DefaultSubscriptionPlan

    for subscription_plan in default_subscription_plans:
        if plan := SubscriptionPlan.objects.filter(name=subscription_plan.name).first():
            subscription_plans[plan.name] = plan
        else:
            subscription_plans[subscription_plan.name] = SubscriptionPlan.objects.create(
                name=subscription_plan.name,
                description=subscription_plan.description,
                price_per_month=subscription_plan.price_per_month,
            )
            logging.info(f"Added SubscriptionPlan {subscription_plan.name}")

    for group in default_usage_plans:
        group_obj, created = PlanFeatureGroup.objects.get_or_create(name=group.name)

        if created:
            logging.info(f"Created group {group.name}")

        item: DefaultFeature
        for item in group.items:
            existing: PlanFeature = PlanFeature.objects.filter(
                slug=item.slug, subscription_plan=SubscriptionPlan.objects.get(name=item.subscription_plan.name)
            ).first()

            if existing:
                description, old_subscription_plan, max_limit_per_month = (
                    existing.description,
                    existing.subscription_plan,
                    existing.max_limit_per_month,
                )

                existing.description = item.description
                existing.max_limit_per_month = item.max_limit_per_month if item.max_limit_per_month != -1 else None

                if (
                    existing.description != description
                    or (existing.max_limit_per_month == None and max_limit_per_month != -1)
                    or (existing.max_limit_per_month != None and max_limit_per_month == -1)
                    or (existing.max_limit_per_month != None and existing.max_limit_per_month != max_limit_per_month)
                ):
                    existing.save()

                    logging.info(f"Updated PlanFeature description/limits for {item.slug}")
            else:
                existing = PlanFeature.objects.create(
                    group=group_obj,
                    description=item.description,
                    slug=item.slug,
                    max_limit_per_month=item.max_limit_per_month,
                    subscription_plan=SubscriptionPlan.objects.get(name=item.subscription_plan.name),
                )
                logging.info(f"Added PlanFeature {item.slug}")



from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.http import require_http_methods

from core.models import UserSettings


@require_http_methods(["POST"])
def update_account_preferences(request):
    currency = request.POST.get("currency", None)
    try:
        usersettings = request.user.user_profile
    except UserSettings.DoesNotExist:
        usersettings = UserSettings.objects.create(user=request.user)

    htmx_return = "base/toasts.html"

    if not request.htmx and not currency:
        return HttpResponse("Invalid Request", status=400)
    elif not currency or currency not in usersettings.CURRENCIES:
        messages.error(request, "Invalid Currency")
        return render(request, htmx_return)

    usersettings.currency = currency

    updated_features: bool = False

    for choice, _ in usersettings.CoreFeatures.choices:
        selected: str | None = request.POST.get(f"selected_{choice}", None)

        if choice in usersettings.disabled_features:  # currently disabled
            if selected:  # enabled
                updated_features = True
                usersettings.disabled_features.remove(choice)
        else:
            if not selected:  # disabled
                updated_features = True
                usersettings.disabled_features.append(choice)

    usersettings.save(update_fields=["disabled_features", "currency"])
    messages.success(request, "Successfully updated preferences")

    if updated_features:
        response = HttpResponse("Success")
        response["HX-Refresh"] = "true"
        return response
    return render(request, htmx_return)



from django.contrib.auth.models import AnonymousUser
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpRequest
from django_htmx.middleware import HtmxDetails

from core.models import User, Organization


class HtmxHttpRequest(HttpRequest):
    htmx: HtmxDetails
    user: User
    no_retarget: bool | None


class UnauthorizedHttpRequest(HttpRequest):
    user: AnonymousUser
    htmx: HtmxDetails
    no_retarget: bool | None


class HtmxAnyHttpRequest(HttpRequest):
    user: User | AnonymousUser
    htmx: HtmxDetails
    no_retarget: bool | None



import logging

import stripe
from django.contrib import messages
from django.db.models import QuerySet
from django.http import HttpResponse
from django.shortcuts import render
from django.urls import reverse

from core.decorators import htmx_only, web_require_scopes
from core.models import User
from core.types.requests import WebRequest
from billing.models import SubscriptionPlan, UserSubscription, StripeCheckoutSession
from billing.service.stripe_customer import get_or_create_customer_id

logger = logging.getLogger(__name__)


@web_require_scopes("billing:manage", api=True, htmx=True)
@htmx_only("billing:dashboard")
def change_plan_endpoint(request: WebRequest):

    plan: SubscriptionPlan | None = None

    if plan_filter := request.POST.get("plan_name"):
        plan = SubscriptionPlan.objects.filter(name=plan_filter).first()
    elif plan_filter := request.POST.get("plan_id"):
        plan = SubscriptionPlan.objects.filter(id=plan_filter).first()

    if not plan:
        messages.error(request, "Invalid plan")
        return render(request, "base/toast.html")
    elif plan.price_per_month == -1 or plan.name.lower() == "enterprise":
        print("THIS PLAN IS ENTERPRISE, currently not implemented")
        messages.error(request, "Invalid plan (not yet implemented)")
        return render(request, "base/toast.html")

    users_plans: QuerySet[UserSubscription] = UserSubscription.filter_by_owner(request.actor)

    if plan.price_per_month == 0 and users_plans.exists():
        messages.error(
            request,
            """
            Unfortunately you have already used up your free trial. Please upgrade to a paid plan to continue.
            If you have another query, feel free to book a call with the project lead + founder!
            <a class="link link-primary" href="https://cal.com/d/88aZVDH9EScpNQnMp4WjEg/myfinances-consumer-call"
                onclick="const e = arguments[0] || window.event; e.stopPropagation();" target=”_blank”>
                <strong>Book here</strong>
            </a>
        """,
        )
        return render(request, "base/toast.html", {"autohide": False})

    users_active_plans: QuerySet[UserSubscription] = users_plans.filter(end_date__isnull=True)
    # if users_active_plans.exists():
    #     for active_plan in users_active_plans:
    #         active_plan.end_date = timezone_now()
    #         active_plan.save()

    line_items = [{"adjustable_quantity": {"enabled": False}, "quantity": 1, "price": plan.stripe_price_id}]  # type: ignore

    checkout_session_django_object = (
        StripeCheckoutSession.objects.create(user=request.actor, plan=plan)
        if isinstance(request.actor, User)
        else StripeCheckoutSession.objects.create(organization=request.actor, plan=plan)
    )

    for feature in plan.features.all():
        if not feature.stripe_price_id:
            continue

        checkout_session_django_object.features.add(feature)

        line_items.append(
            {
                # "adjustable_quantity": {
                #     "enabled": False
                # },
                "price": feature.stripe_price_id,
                # "quantity": 1,
            }
        )

    customer_id = get_or_create_customer_id(request.actor)

    if isinstance(request.actor, User):
        customer_email = request.actor.email
    else:
        customer_email = request.actor.leader.email

    checkout_session = stripe.checkout.Session.create(
        customer=customer_id,
        customer_email=customer_email if not customer_email else None,  # type: ignore[arg-type]
        line_items=line_items,  # type: ignore[arg-type]
        mode="subscription",
        # return_url="http://127.0.0.1:8000" + reverse("billing:stripe_checkout_failed_response"),
        cancel_url=request.build_absolute_uri(reverse("billing:dashboard")),
        success_url=request.build_absolute_uri(reverse("billing:stripe_checkout_success_response")),
        metadata={"dj_checkout_uuid": checkout_session_django_object.uuid, "dj_subscription_plan_id": str(plan.id)},
        saved_payment_method_options={"payment_method_save": "enabled"},
    )

    checkout_session_django_object.stripe_session_id = checkout_session.id

    checkout_session_django_object.save()

    # UserSubscription.objects.create(owner=request.actor, subscription_plan=plan)
    messages.success(request, "Great! Redirecting you to stripe now!")
    r = HttpResponse(status=200)
    r["HX-Redirect"] = str(checkout_session.url)
    return r



from dataclasses import dataclass
from typing import TypeVar, Any, Optional, Generic

T = TypeVar("T")


def extract_to_dataclass(request, class_type: T, request_types: list[str], *args, **kwargs) -> T:
    """

    Turn kwargs from Key:Value and get request.POST.get(value) and set class.key = request.POST.get(value)

    Usage:

    from pydantic.dataclasses import dataclass
    from typing import Optional

    @dataclass
    class MyView:
        name: str
        age: int
        different: bool
        non_required: Optional[str]

    def myview(request):
        try:
            data = extract_to_dataclass(request, MyView, ["post"], "name", "age", diff_bool="different")
        except pydantic.ValidationError:
            pass
    """
    data: dict = {}
    if "get" in request_types:
        if args:
            data |= {key: request.GET.get(key) for key in args}

        if kwargs:
            data |= {key: request.GET.get(value) for key, value in kwargs.items()}

    if "post" in request_types:
        if args:
            data |= {key: request.POST.get(key) for key in args}

        if kwargs:
            data |= {key: request.POST.get(value) for key, value in kwargs.items()}

    if "headers" in request_types:
        if args:
            data |= {key: request.headers.get(key) for key in args}

        if kwargs:
            data |= {key: request.headers.get(value) for key, value in kwargs.items()}

    if isinstance(class_type, type):
        return class_type(**data)
    else:
        raise TypeError("class_type must be a class")


class BaseServiceResponse(Generic[T]):
    _success: bool = False
    _response: Optional[T] = None
    _error_message: str = ""
    _status_code: Optional[int] = None

    def __init__(self, success: bool = False, response: Optional[T] = None, error_message: str = "", status_code: Optional[int] = None):
        self._success = success
        self._response = response
        self._error_message = error_message
        self._status_code = status_code

    @property
    def success(self) -> bool:
        if not isinstance(self._success, bool):
            raise TypeError("success must be a boolean")

        return self._success

    @property
    def response(self) -> T:
        if self._response is None:
            raise TypeError("response must be present if it was a successful response")
        return self._response

    @property
    def error_message(self) -> str:
        if not isinstance(self._error_message, str):
            raise TypeError("error_message must be a string")
        return self._error_message

    @property
    def status_code(self) -> int:
        if not isinstance(self._status_code, int):
            raise TypeError("status code must be an integer")
        return self._status_code

    @property
    def failed(self) -> bool:
        return not self.success

    @property
    def error(self) -> str:
        return self.error_message if self.failed else "Unknown error"

    def __post_init__(self):
        if self.success and self.response is None:
            raise ValueError("Response cannot be None when success is True.")
        if not self.success and self.response is not None:
            raise ValueError("Response must be None when success is False.")
        if not self.success and not self.error_message:
            raise ValueError("Error message cannot be empty when success is False.")


# * BaseServiceResponse Usage

# from backend.utils.dataclasses import BaseServiceResponse
#
#
# class XyzServiceResponse(BaseServiceResponse[ResponseObject]):
#     response: Optional[ResponseObject] = None
# or
# ...

# * Return Response

# return CreateClientServiceResponse(False, error_message="my error")
# return CreateClientServiceResponse(False, ClientObject)

# * View Usage
#
# client_response: CreateClientServiceResponse = create_client(request)
#
# if client_response.failed:
#     print(client_response.error)
# else:
#     print(client_response.response) # < ClientObject>



from functools import wraps
from rest_framework import status

from core.models import TeamMemberPermission
from core.api.public.helpers.response import APIResponse

import logging

logger = logging.getLogger(__name__)


def require_scopes(scopes):
    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            token = request.auth
            if not token:
                logger.info(
                    f"Authentication credentials were not provided in api request |" f" {request.META.get('REMOTE_ADDR', 'Unknown IP')}"
                )
                return APIResponse(False, {"detail": "Authentication credentials were not provided."}, status=status.HTTP_401_UNAUTHORIZED)

            if request.team_id and not request.team:
                return APIResponse(False, {"detail": "Team not found."}, status=status.HTTP_404_NOT_FOUND)

            if request.team:
                # Check for team permissions based on team_id and scopes
                if not request.team.is_owner(token.user) and not request.team.is_logged_in_as_team(request):
                    team_permissions = TeamMemberPermission.objects.filter(team=request.team, user=token.user).first()
                    if not team_permissions or not all(scope in team_permissions.scopes for scope in scopes):
                        return APIResponse(False, {"detail": "Permission denied."}, status=status.HTTP_403_FORBIDDEN)

            # Check for global API Key permissions based on token scopes
            if not all(scope in token.scopes for scope in scopes):
                return APIResponse(False, {"detail": "Permission denied."}, status=status.HTTP_403_FORBIDDEN)

            token.update_last_used()

            return view_func(request, *args, **kwargs)

        _wrapped_view.required_scopes = scopes
        return _wrapped_view

    return decorator



from django.http import HttpResponseRedirect


def redirect_to_last_visited(request, fallback_url="dashboard"):
    """
    Redirects user to the last visited URL stored in session.
    If no previous URL is found, redirects to the fallback URL.
    :param request: HttpRequest object
    :param fallback_url: URL to redirect to if no previous URL found
    :return: HttpResponseRedirect object
    """
    try:
        last_visited_url = request.session.get("last_visited", fallback_url)
        return HttpResponseRedirect(last_visited_url)
    except KeyError:
        return HttpResponseRedirect(fallback_url)



from django.contrib import messages
from django.shortcuts import render
from django.views.decorators.http import require_POST

from core.decorators import has_entitlements
from core.models import Organization
from core.types.htmx import HtmxHttpRequest


@require_POST
@has_entitlements("organizations")
# @quota_usage_check("teams-count", api=True, htmx=True)
def create_team(request: HtmxHttpRequest):
    name = request.POST.get("name")

    if not name:
        messages.error(request, "A team name field must be filled.")
        return render(request, "partials/messages_list.html")

    if Organization.objects.filter(name=name).exists():
        messages.error(request, "A team with this name already exists.")
        return render(request, "partials/messages_list.html")

    team = Organization.objects.create(name=name, leader=request.user)

    # QuotaUsage.create_str(request.user, "teams-count", team.id)
    # QuotaUsage.create_str(request.user, "teams-joined", team.id)

    if not request.user.logged_in_as_team:
        request.user.logged_in_as_team = team
        request.user.save()

    messages.success(request, f"Successfully created team {name} with the ID of #{team.id}")
    response = render(request, "partials/messages_list.html")
    response["HX-Refresh"] = "true"
    return response



import re

from django.template import Library, Node
from django.utils.encoding import force_str

register = Library()


def strip_spaces_in_tags(value):
    value = force_str(value)
    value = re.sub(r"\s+", " ", value)
    value = re.sub(r">\s+", ">", value)
    value = re.sub(r"\s+<", "<", value)
    return value


class NoSpacesNode(Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        return strip_spaces_in_tags(self.nodelist.render(context).strip())


@register.tag
def nospaces(parser, token):
    """
    Removes any duplicite whitespace in tags and text. Can be used as supplementary tag for {% spaceless %}::

        {% nospaces %}
        <strong>
            Hello
            this is text
        </strong>
        {% nospaces %}

    Returns::
        <strong>Hello this is text</strong>
    """
    nodelist = parser.parse(("endnospaces",))
    parser.delete_first_token()
    return NoSpacesNode(nodelist)



from dataclasses import dataclass
from typing import Union

from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import redirect, render
from django.views.decorators.http import require_http_methods

from core.decorators import superuser_only
from core.models import QuotaIncreaseRequest, QuotaLimit, QuotaUsage, QuotaOverrides
from core.types.htmx import HtmxHttpRequest


# from backend.utils.quota_limit_ops import quota_usage_check_under


def submit_request(request: HtmxHttpRequest, slug) -> HttpResponse:
    if not request.htmx:
        return redirect("quotas")

    new_value = request.POST.get("new_value", "")
    reason = request.POST.get("reason", "")

    try:
        quota_limit = QuotaLimit.objects.get(slug=slug)
    except QuotaLimit.DoesNotExist:
        return error(request, "Failed to get the quota limit type")

    # usage_per_item = quota_usage_check_under(request, "quota_increase-request", extra_data=quota_limit.id, api=True, htmx=True)
    # usage_per_month = quota_usage_check_under(
    #     request, "quota_increase-requests_per_month_per_quota", extra_data=quota_limit.id, api=True, htmx=True
    # )

    # if not isinstance(usage_per_item, bool):
    #     return usage_per_item

    # if not isinstance(usage_per_month, bool):
    #     return usage_per_month

    current = quota_limit.get_quota_limit(request.user)

    validate = validate_request(new_value, reason, current)

    if isinstance(validate, Error):
        return error(request, validate.message)

    quota_increase_request = QuotaIncreaseRequest.objects.create(
        user=request.user, quota_limit=quota_limit, new_value=new_value, current_value=current, reason=reason
    )

    QuotaUsage.create_str(request.user, "quota_increase-request", quota_increase_request.id)
    QuotaUsage.create_str(request.user, "quota_increase-requests_per_month_per_quota", quota_limit.id)

    messages.success(request, "Successfully submitted a quota increase request")
    return render(request, "base/toast.html")


@dataclass
class Error:
    message: str


def error(request: HtmxHttpRequest, message: str) -> HttpResponse:
    messages.error(request, message)
    return render(request, "partials/messages_list.html")


def validate_request(new_value, reason, current) -> Union[bool, Error]:
    if not new_value:
        return Error("Please enter a valid increase value")

    try:
        new_value = int(new_value)
        if new_value <= current:
            raise ValueError
    except ValueError:
        return Error("Please enter a valid increase value that is above your current limit.")

    if len(reason) < 25:
        return Error("Please enter a valid reason for the increase.")

    return True


@superuser_only
@require_http_methods(["DELETE", "POST"])
def approve_request(request: HtmxHttpRequest, request_id) -> HttpResponse:
    if not request.htmx:
        return redirect("quotas")
    try:
        quota_request = QuotaIncreaseRequest.objects.get(id=request_id)
    except QuotaIncreaseRequest.DoesNotExist:
        return error(request, "Failed to get the quota increase request")

    try:
        quota_override_existing = QuotaOverrides.objects.get(user=quota_request.user, quota_limit=quota_request.quota_limit)
        quota_override_existing.value = quota_request.new_value
        quota_override_existing.save()
    except QuotaOverrides.DoesNotExist:
        QuotaOverrides.objects.create(
            user=quota_request.user,
            value=quota_request.new_value,
            quota_limit=quota_request.quota_limit,
        )

    quota_limit_for_increase = QuotaLimit.objects.get(slug="quota_increase-request")
    QuotaUsage.objects.filter(user=quota_request.user, quota_limit=quota_limit_for_increase, extra_data=quota_request.id).delete()
    quota_request.status = "approved"
    quota_request.save()

    try:
        QuotaUsage.objects.get(
            quota_limit=QuotaLimit.objects.get(slug="quota_increase-requests_per_month_per_quota"), extra_data=quota_request.quota_limit_id
        ).delete()
    except QuotaUsage.DoesNotExist:
        ...

    return HttpResponse(status=200)


@superuser_only
@require_http_methods(["DELETE", "POST"])
def decline_request(request: HtmxHttpRequest, request_id) -> HttpResponse:
    if not request.htmx:
        return redirect("quotas")
    try:
        quota_request = QuotaIncreaseRequest.objects.get(id=request_id)
    except QuotaIncreaseRequest.DoesNotExist:
        return error(request, "Failed to get the quota increase request")

    quota_limit_for_increase = QuotaLimit.objects.get(slug="quota_increase-request")
    QuotaUsage.objects.filter(user=quota_request.user, quota_limit=quota_limit_for_increase, extra_data=quota_request.id).delete()
    quota_request.status = "decline"
    quota_request.save()

    return HttpResponse(status=200)



from django.urls import path

from . import healthcheck

urlpatterns = [
    path(
        "ping/",
        healthcheck.ping,
        name="ping",
    ),
    path(
        "healthcheck/",
        healthcheck.healthcheck,
        name="healthcheck",
    ),
]

app_name = "healthcheck"
